# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import bz2
import datetime as dt
import matplotlib.pyplot as plt
import pytest
import warnings

import pydarn

with bz2.open('test/data/test.fitacf.bz2') as fp:
    fitacf_stream = fp.read()
data = pydarn.SuperDARNRead(fitacf_stream, True).read_fitacf()


class TestRTP_defaults:

    def test_range_time_defaults(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_range_time(data)

    def test_coord_time_defaults(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_coord_time(data)

    def test_normal_time_series(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_time_series(data)

    def test_normal_summary_plot(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_summary(data)


@pytest.mark.parametrize('background', ['w'])
@pytest.mark.parametrize('zmin', [0, -200])
@pytest.mark.parametrize('zmax', [200, 1000])
@pytest.mark.parametrize('groundscatter_params', [True, 'grey'])
@pytest.mark.parametrize('colorbar_label', ['test'])
@pytest.mark.parametrize('yspacing', [150, 250])
@pytest.mark.parametrize('range_estimation',
                         [pydarn.RangeEstimation.RANGE_GATE,
                          pydarn.RangeEstimation.GSMR,
                          pydarn.RangeEstimation.HALF_SLANT])
@pytest.mark.parametrize('ymin', [10])
@pytest.mark.parametrize('ymax', [70])
@pytest.mark.parametrize('parameters', ['v', 'p_l', 'w_l'])
@pytest.mark.parametrize('cmap', [plt.get_cmap('rainbow')])
@pytest.mark.parametrize('start_time', [dt.datetime(2018, 4, 4, 6, 2)])
@pytest.mark.parametrize('end_time', [dt.datetime(2018, 4, 4, 6, 4)])
@pytest.mark.parametrize('date_fmt', ['%H:%M'])
@pytest.mark.parametrize('round_start', [False])
class TestRangTime:

    def test_parameters_range_time(self, groundscatter_params,
                                   parameters, background, zmin, zmax,
                                   start_time, end_time, ymin, ymax,
                                   colorbar_label, yspacing, round_start,
                                   range_estimation, cmap, date_fmt):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_range_time(data, beam_num=15,
                                       parameter=parameters, channel=1,
                                       groundscatter=groundscatter_params,
                                       background=background, zmin=zmin,
                                       zmax=zmax, start_time=start_time,
                                       end_time=end_time, ymin=ymin, ymax=ymax,
                                       colorbar_label=colorbar_label,
                                       yspacing=yspacing,
                                       round_start=round_start,
                                       range_estimation=range_estimation,
                                       cmap=cmap, date_fmt=date_fmt)
        plt.close('all')

    def test_parameters_channel_2_range_time(self, groundscatter_params,
                                             parameters, background, zmin,
                                             zmax, start_time, end_time,
                                             ymin, ymax, colorbar_label,
                                             yspacing, round_start,
                                             range_estimation,
                                             cmap, date_fmt):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_range_time(data, beam_num=9,
                                       parameter=parameters, channel=2,
                                       groundscatter=groundscatter_params,
                                       background=background, zmin=zmin,
                                       zmax=zmax, start_time=start_time,
                                       end_time=end_time, ymin=ymin, ymax=ymax,
                                       colorbar_label=colorbar_label,
                                       yspacing=yspacing,
                                       round_start=round_start,
                                       range_estimation=range_estimation,
                                       cmap=cmap, date_fmt=date_fmt)
        plt.close('all')


@pytest.mark.parametrize('parameters_scalar', ['tfreq', 'cp', 'nave',
                                               'p_l', 'w_l', 'v'])
@pytest.mark.parametrize('gate', [38, 48])
@pytest.mark.parametrize('scale', ['linear', 'log'])
@pytest.mark.parametrize('cp_name', [True, False])
@pytest.mark.parametrize('color', ['green'])
@pytest.mark.parametrize('linestyle', ['--'])
@pytest.mark.parametrize('linewidth', [0.5, 2])
@pytest.mark.parametrize('start_time', [dt.datetime(2018, 4, 4, 6, 2)])
@pytest.mark.parametrize('end_time', [dt.datetime(2018, 4, 4, 6, 4)])
@pytest.mark.parametrize('date_fmt', ['%H:%M'])
@pytest.mark.parametrize('round_start', [False])
class TestTimeSeries:

    def test_parameters_time_series_channel2(self, parameters_scalar, gate,
                                             scale, cp_name, color, linestyle,
                                             linewidth, date_fmt, end_time,
                                             start_time, round_start):
        """

        """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_time_series(data, parameter=parameters_scalar,
                                        beam_num=9, gate=gate,
                                        start_time=start_time,
                                        end_time=end_time, date_fmt=date_fmt,
                                        channel=2, scale=scale,
                                        cp_name=cp_name, color=color,
                                        linestyle=linestyle,
                                        linewidth=linewidth,
                                        round_start=round_start,)

    def test_parameters_time_series_channel1(self, parameters_scalar, gate,
                                             scale, cp_name, color, linestyle,
                                             linewidth, date_fmt, end_time,
                                             start_time, round_start):
        """

        """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_time_series(data, parameter=parameters_scalar,
                                        beam_num=15, gate=gate,
                                        start_time=start_time,
                                        end_time=end_time, date_fmt=date_fmt,
                                        channel=1, scale=scale,
                                        cp_name=cp_name, color=color,
                                        linestyle=linestyle,
                                        linewidth=linewidth,
                                        round_start=round_start,)


@pytest.mark.parametrize('fig_size', [(11, 10), (10, 11)])
@pytest.mark.parametrize('watermark', [False, True])
@pytest.mark.parametrize('boundary', [{'w_l': (0, 100),
                                       'nave': (0, 30),
                                       'elv': (0, 20),
                                       'tfreq': (0, 15)}])
@pytest.mark.parametrize('cmaps', [{'elv':'rainbow',
                                    'p_l': 'rainbow',
                                    'vel': 'rainbow'}])
@pytest.mark.parametrize('lines', [{'nave': '--', 'tfreq': '-'}])
@pytest.mark.parametrize('plot_elv', [False])
@pytest.mark.parametrize('title', ['test test'])
@pytest.mark.parametrize('background', ['grey'])
@pytest.mark.parametrize('groundscatter', [False, 'grey'])
@pytest.mark.parametrize('range_estimation',
                         [pydarn.RangeEstimation.RANGE_GATE,
                          pydarn.RangeEstimation.GSMR,
                          pydarn.RangeEstimation.HALF_SLANT])
class TestSummaryPlots:

    def test_params_summary_plots(self, fig_size, watermark, boundary, cmaps,
                                  lines, plot_elv, title, background,
                                  groundscatter, range_estimation):
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_summary(data, beam_num=15, channel=1,
                                    figsize=fig_size, watermark=watermark,
                                    boundary=boundary, cmaps=cmaps,
                                    lines=lines, plot_elv=plot_elv,
                                    title=title, background=background,
                                    groundscatter=groundscatter,
                                    range_estimation=range_estimation)
        plt.close('all')

    def test_beam9_channel2_summary_plots(self, fig_size, watermark, boundary,
                                          cmaps, lines, plot_elv, title,
                                          background, groundscatter,
                                          range_estimation):
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_summary(data, beam_num=9, channel=2,
                                    figsize=fig_size, watermark=watermark,
                                    boundary=boundary, cmaps=cmaps,
                                    lines=lines, plot_elv=plot_elv,
                                    title=title, background=background,
                                    groundscatter=groundscatter,
                                    range_estimation=range_estimation)
        plt.close('all')


@pytest.mark.parametrize('background', ['w'])
@pytest.mark.parametrize('zmin', [0, -200])
@pytest.mark.parametrize('zmax', [200, 1000])
@pytest.mark.parametrize('groundscatter_params', [True, False])
@pytest.mark.parametrize('colorbar_label', ['test'])
@pytest.mark.parametrize('yspacing', [150, 250])
@pytest.mark.parametrize('range_estimation', [
                          pydarn.RangeEstimation.SLANT_RANGE,
                          pydarn.RangeEstimation.HALF_SLANT])
@pytest.mark.parametrize('ymin', [10])
@pytest.mark.parametrize('ymax', [70])
@pytest.mark.parametrize('parameters', ['v', 'p_l', 'w_l'])
@pytest.mark.parametrize('cmap', ['viridis'])
@pytest.mark.parametrize('start_time', [dt.datetime(2018, 4, 4, 6, 2)])
@pytest.mark.parametrize('end_time', [dt.datetime(2018, 4, 4, 6, 4)])
@pytest.mark.parametrize('date_fmt', ['%H:%M'])
@pytest.mark.parametrize('round_start', [False])
@pytest.mark.parametrize('latlon', ['lat', 'lon'])
@pytest.mark.parametrize('plot_equatorward', [False])
class TestCoordTime:

    def test_parameters_coord_time(self, groundscatter_params,
                                   parameters, background, zmin, zmax,
                                   start_time, end_time, ymin, ymax,
                                   colorbar_label, yspacing, round_start,
                                   range_estimation, cmap, date_fmt,
                                   plot_equatorward, latlon):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.RTP.plot_coord_time(data, beam_num=15,
                                       parameter=parameters, channel=1,
                                       groundscatter=groundscatter_params,
                                       background=background, zmin=zmin,
                                       zmax=zmax, start_time=start_time,
                                       end_time=end_time, ymin=ymin, ymax=ymax,
                                       colorbar_label=colorbar_label,
                                       yspacing=yspacing,
                                       round_start=round_start,
                                       range_estimation=range_estimation,
                                       cmap=cmap, date_fmt=date_fmt,
                                       plot_equatorward=plot_equatorward,
                                       latlon=latlon)
        plt.close('all')


# Copyright (C) 2019 SuperDARN Canada, University Saskatchewan
# Author: Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modification:
#
"""
This module contains SuperDARN radar modes classified by cpid number
Current documentation on modes:
    http://superdarn.thayer.dartmouth.edu/info/modes.html
"""
# TODO: find/create better radar mode/frequency bands documentation


class SuperDARNCpids():

    cpids = {26401: 'stereoscan',
             26009: 'stereoscan',
             26008: 'stereoscan',
             26007: 'stereoscan',
             26006: 'stereoscan',
             26005: 'stereoscan',
             26004: 'stereoscan',
             26003: 'stereoscan',
             26002: 'stereoscan',
             6401: 'stereoscan',
             150: 'normalscan slow',
             151: 'normalscan fast',
             152: 'stereoscan',
             153: 'stereoscan',
             155: 'normalsound slow',
             157: 'normalsound fast',
             200: 'rbspscan',
             3200: 'risrscan',
             3250: 'twotsg slow',
             3251: 'twotsg fast',
             3252: 'twotsg slow',
             3253: 'twotsg fast',
             3300: 'themisscan',
             3333: 'ddstest',
             3350: 'ulfscan',
             3370: 'epopsound',
             3375: 'longsound',
             3380: 'politescan',
             3400: 'fivepulse',
             3450: 'heatsound',
             3500: 'twofsound slow',
             3501: 'twofsound fast',
             3502: 'twofsound slow 7-8 pulses',
             3503: 'twofsound fast 7-8 pulses',
             3505: 'twofsound',
             3517: 'g10scan',
             3520: 'uafsound slow',
             3521: 'uafsound fast',
             3550: 'twofonebm',
             3560: 'multifonebm',
             3600: 'tauscan slow',
             3601: 'tauscan fast',
             8510: 'ltuseqscan slow',
             8511: 'ltuseqscan fast',
             9211: 'pcpscan',
             9151: 'pcodecamp'}



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author(s): Daniel Billett, Marina Schmidt
#
# Copyright (C) 2012 Johns Hopkins University/Applied Physics Laboratory
# Author: K.Baker, R.J.Barnes & D. Andre
#
# The following methods:
# - geographic_cell_positions
# - geocentric_coordinates
# - cell_geocentric_coordinates
# - geocentric2flattening
# - geodetic2geocentric
# - geocentric2geodetic
# was copied from
# https://github.com/SuperDARN/rst/superdarn/src.lib/tk/rpos.1.7/src/cnvtcoord.c
# and converted to python for usage in pyDARN
#
# Modifications:
# 2022-03-10 MTS added kwargs to the functions just to avoid errors when
#                extra keys are passed in from other functions
# 2023-01-03 CJM added functions to calculate azimuth from XS and SC existing
#                codebase, can be expanded and added to when required
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
"""
This module focuses on geographic calculations used for converting gate
information to geographic location
"""
import numpy as np
from pydarn import VHModels, EARTH_EQUATORIAL_RADIUS


def geocentric_coordinates(target_range: float, psi: float, boresight: float,
                           virtual_height_model: VHModels =
                           VHModels.STANDARD,
                           **kwargs):
    """
    Calculates the geocentric coordinates of gate cell  point,
    using either the standard or Chisham virtual height model.

    Parameters
    ----------
        radar_lat : float
            radars site latitude [rad]
        radar_lon : float
            radars site longitude [lon]
        target_range: float
            The range from the instrument to the target (echo) [km]
        cell_height : float
            virtual height of the gate cell [km]
        psi: int
            [rad]
        boresight: float
            boresight of the radar beam [rad]
        virtual_height_model: VHModels
            use for choosing type of virtual height
            default: VHModels.STANDARD

    Returns
    -------
        cell_lat: float
            latitude of the range gate in geographic coordinates [rad]
        cell_lon: float
            longitude of the range gate in geographic coordinates [rad]

    """
    """
    Mapping ionospheric backscatter measured by the SuperDARN HF
    radars – Part 1: A new empirical virtual height model by
    G. Chisham 2008 (https://doi.org/10.5194/angeo-26-823-2008)
    """
    x_height = virtual_height_model(target_range=target_range, **kwargs)

    # calculate the radius over the earth underneath
    # the radar and range gate cell
    rlat, rlon, r_radar, delta = geodetic2geocentric(**kwargs)
    r_cell = r_radar

    psi_cos_2 = np.cos(psi)**2
    psi_sin_2 = np.sin(psi)**2

    while_flag = True
    while while_flag:
        # distance between the gate cell to the earth's centre [km]
        cell_rho = r_cell + x_height
        # elevation angle relative to local horizon [rad]
        rel_elv = np.arcsin(((cell_rho**2) - (r_radar**2) - target_range**2) /
                            (2.0 * r_radar * target_range))
        # estimate elevation for multi-hop propagation
        if virtual_height_model == VHModels.CHISHAM and target_range > 2137.5:
            gamma = np.arccos((r_radar**2 + cell_rho**2 - target_range**2) /
                              (2.0 * r_radar * cell_rho))
            beta = np.arcsin(r_radar * np.sin(gamma/3.0) /
                             (target_range/3.0))
            # Elevation angle used for estimating off-array normal
            # azimuth [rad]
            xelv = (np.pi/2) - beta - (gamma/3.0)
        else:
            xelv = rel_elv

        # Estimate the off-array-normal azimuth in radians
        elv_sin_2 = np.sin(xelv)**2

        est_azimuth = psi_cos_2 - elv_sin_2
        if est_azimuth < 0:
            tan_azimuth = 1e32
        else:
            # in radians
            tan_azimuth = np.sqrt(psi_sin_2 /
                                  (psi_cos_2 - elv_sin_2))
        # azimuth in [rad]
        if psi > 0:
            azimuth = np.arctan(tan_azimuth)
        else:
            azimuth = -np.arctan(tan_azimuth)

        # azimuth of the gate cell [rad]
        cell_azimuth = azimuth + boresight
        flatten_azimuth = geocentric2flattening(delta=delta,
                                                azimuth=cell_azimuth,
                                                elv=xelv)
        cell_rho, cell_lat, cell_lon = \
            cell_geocentric_coordinates(lat=rlat, lon=rlon,
                                        rho=r_radar,
                                        azimuth=flatten_azimuth,
                                        elv=rel_elv,
                                        r=target_range)

        # recalculate the radius under the gate cell and centre of earth
        r_cell = geocentric2geodetic(lat=cell_lat, lon=cell_lon)
        cell_heightx = cell_rho - r_cell
        # this ensures convergence on the cell point
        while_flag = abs(cell_heightx - x_height) > 0.5

    return cell_lat, cell_lon


# fldpnt
def cell_geocentric_coordinates(lat: float, lon: float, rho: float,
                                azimuth: float, elv: float, r: float,
                                **kwargs):
    """
    Calculates the geocentric coordinates of a gate cell point given the
    angular geocentric coordinates of the point of origin,
    the azimuth, elevation, and slant range.

    parameters
    ----------
        lat: float
            geocentric latitude [rad] of the radar site
        lon : float
            geocentric longitude [rad] of the radar site
        rho: float
            radius from the radar site to the centre of the earth [km]
        azimuth: float
            azimuth with corrected earth flattening in [rad] for the cell
        elv: float
            relative elevation angle [rad]
        r: float
            slant range of the cell [km]

    returns
    -------
        rho: float
            radius from cell point to centre of the earth [km]
        lat: float
            geocentric cell latitude [rad]
        lon: float
            geocentric cell longitude [rad]
    """
    cos_lat = np.cos(np.pi/2 - lat)
    sin_lat = np.sin(np.pi/2 - lat)

    cos_lon = np.cos(lon)
    sin_lon = np.sin(lon)

    cos_azimuth = np.cos(azimuth)
    sin_azimuth = np.sin(azimuth)

    sin_elv = np.sin(elv)
    cos_elv = np.cos(elv)

    # earth centered spherical coordinates [km*rad]
    rx = rho * sin_lat * cos_lon
    ry = rho * sin_lat * sin_lon
    rz = rho * cos_lat

    # earth local Cartesian (x-south, y-east, z-up)

    # [km*rad]
    local_x = -r * cos_elv * cos_azimuth
    local_y = r * cos_elv * sin_azimuth
    local_z = r * sin_elv

    # convert to global Cartesian
    global_x = cos_lat * local_x + sin_lat * local_z
    global_y = local_y
    global_z = -sin_lat * local_x + cos_lat * local_z

    local_x = cos_lon * global_x - sin_lon * global_y
    local_y = sin_lon * global_x + cos_lon * global_y
    local_z = global_z

    # find the global Cartesian coordinates for the new point
    global_x = rx + local_x
    global_y = ry + local_y
    global_z = rz + local_z

    # convert Cartesian back to spherical
    rho = np.sqrt(global_x**2 + global_y**2 + global_z**2)
    lat = np.pi/2 - np.arccos(global_z/rho)
    if global_x == 0 and global_y == 0:
        lon = 0
    else:
        lon = np.arctan2(global_y, global_x)

    return rho, lat, lon


# goecnvrt
def geocentric2flattening(delta: float, azimuth: float, elv: float, **kwargs):
    """
    Adjust azimuth for the oblateness of the Earth

    Parameters
    ----------
        delta: float
            distance before geodetic and geocentric latitude [rad]
        azimuth: float
            azimuth of geocentric cell point [rad]
        elv: float
            elevation angle [rad]

    Returns
    -------
        azimuth_flattening: float
            adjusted azimuth due the earths oblateness

    """
    cos_elv = np.cos(elv)
    sin_elv = np.sin(elv)

    cos_azimuth = np.cos(azimuth)
    sin_azimuth = np.sin(azimuth)

    cos_delta = np.cos(delta)
    sin_delta = np.sin(delta)

    kxg = cos_elv * sin_azimuth
    kyg = cos_elv * cos_azimuth
    kzg = sin_elv

    kxr = kxg
    kyr = kyg * cos_delta + kzg * sin_delta

    azimuth_flattening = np.arctan2(kxr, kyr)

    return azimuth_flattening


# geodtgc, iopt > 0
def geodetic2geocentric(lat: float, lon: float, **kwargs):
    """
    convert geodetic coordinates to geocentric
    radian version

    parameters
    ----------
        lat: float
            geodetic latitude [rad]
        lon: float
            geodetic longitude [rad]
    returns
    -------
        glat: float
            geocentric latitude [rad]
        glon: float
            geocentric longitude [rad]
        rho: float
            distance from the latitude to the centre of the earth [km]
        delta: float
            difference between geodetic and geocentric latitude [rad]
    """
    # WGS 84 oblate spheroid defining parameters
    # reciprocal flattening
    f = 1.0 / 298.257223563
    # b is in [km] semi minor axis of earth
    b = EARTH_EQUATORIAL_RADIUS * (1.0 - f)
    # e2 is the ellipticity
    e2 = EARTH_EQUATORIAL_RADIUS**2 / b**2 - 1

    glat = np.arctan(b**2 / EARTH_EQUATORIAL_RADIUS**2 * np.tan(lat))
    # glon [rad]
    glon = lon

    if glon > np.pi:
        glon = glon - 2 * np.pi
    # grho is km?
    rho = EARTH_EQUATORIAL_RADIUS / np.sqrt(1 + e2 * np.sin(glat)**2)
    # delta in [rad]
    delta = lat - glat
    return glat, glon, rho, delta


# geodtgc, iopt < 0
def geocentric2geodetic(lat: float, lon: float, **kwargs):
    """
    convert geocentric coordinates to geodetic
    radian version

    parameters
    ----------
        lat: float
            geocentric latitude [rad]
        lon: float
            geocentric longitude [rad]
    returns
    -------
        dlat: float
            geodetic latitude [rad]
        dlon: float
            geodetic longitude [rad]
        rho: float
            distance from the latitude to the centre of the earth [km]
        delta: float
            difference between geocentric and geodetic latitude [rad]
    """
    # WGS 84 oblate spheroid defining parameters
    f = 1.0 / 298.257223563
    b = EARTH_EQUATORIAL_RADIUS * (1.0 - f)
    e2 = EARTH_EQUATORIAL_RADIUS**2 / b**2 - 1

    rho = EARTH_EQUATORIAL_RADIUS / np.sqrt(1+e2*np.sin(lat)**2)

    return rho


def geod2geoc(lat, lon, inverse=False):
    """Converts position from geodetic to geocentric or vice-versa.
    Based on the IAU 1964 oblate spheroid model of the Earth.
    Parameters
    ----------
    lat : float
        latitude [degree]
    lon : float
        longitude [degree]
    inverse : Optional[bool]
        inverse conversion (geocentric to geodetic).  Default is false.
    Returns
    -------
    lat_out : float
        latitude [degree] (geocentric/detic if inverse=False/True)
    lon_out : float
        longitude [degree] (geocentric/detic if inverse=False/True)
    rade : float
        Earth radius [km] (geocentric/detic if inverse=False/True)
    """
    a = 6378.16
    f = 1.0 / 298.25
    b = a * (1.0 - f)
    e2 = (a**2 / b**2) - 1.0

    if not inverse:
        # geodetic into geocentric
        lat_out = np.degrees(np.arctan(b**2 / a**2 * np.tan(np.radians(lat))))
        lon_out = lon
    else:
        # geocentric into geodetic
        lat_out = np.degrees(np.arctan(a**2 / b**2 * np.tan(np.radians(lat))))
        lon_out = lon

    rade = a / np.sqrt(1.0 + e2 * np.sin(np.radians(lat_out)) ** 2)

    return lat_out, lon_out, rade


def globalspherical2globalcartesian(xin, yin, zin, inverse=False):
    """Converts a position from global spherical (geocentric) to global
    cartesian (and vice-versa).

    Parameters
    ----------
    xin : float
        latitude [degree] or global cartesian X [km]
    yin : float
        longitude [degree] or global cartesian Y [km]
    zin : float
        distance from center of the Earth [km] or global cartesian Z [km]
    inverse : Optional[bool]
        inverse conversion

    Returns
    -------
    xout : float
        global cartesian X [km] (inverse=False) or latitude [degree]
    yout : float
        global cartesian Y [km] (inverse=False) or longitude [degree]
    zout : float
        global cartesian Z [km] (inverse=False) or distance from the center of
        the Earth [km]

    Notes
    -------
    The global cartesian coordinate system is defined as:
        - origin: center of the Earth
        - x-axis in the equatorial plane and through the prime meridian.
        - z-axis in the direction of the rotational axis and through the North
          pole
    The meaning of the input (x,y,z) depends on the direction of the conversion
    (to global cartesian or to global spherical).
    """

    if not inverse:
        # Global spherical to global cartesian
        xout = zin * np.cos(np.radians(xin)) * np.cos(np.radians(yin))
        yout = zin * np.cos(np.radians(xin)) * np.sin(np.radians(yin))
        zout = zin * np.sin(np.radians(xin))
    else:
        # Calculate latitude (xout), longitude (yout) and distance from center
        # of the Earth (zout)
        zout = np.sqrt(xin**2 + yin**2 + zin**2)
        xout = np.degrees(np.arcsin(zin / zout))
        yout = np.degrees(np.arctan2(yin, xin))

    return xout, yout, zout


def globalcartesian2localcartesian(X, Y, Z, lat, lon, rho, inverse=False):
    """Converts a position from global cartesian to local cartesian
    (or vice-versa).

    Parameters
    ----------
    X : float
        global cartesian X [km] or local cartesian X [km]
    Y : flaot
        global cartesian Y [km] or local cartesian Y [km]
    Z : float
        global cartesian Z [km] or local cartesian Z [km]
    lat : float
        geocentric latitude [degree] of local cartesian system origin
    lon : float
        geocentric longitude [degree] of local cartesian system origin
    rho : float
        distance from center of the Earth [km] of local cartesian system origin
    inverse : Optional[bool]
        inverse conversion

    Returns
    -------
    X : float
        local cartesian X [km] or global cartesian X [km]
    Y : float
        local cartesian Y [km] or global cartesian Y [km]
    Z : float
        local cartesian Z [km] or global cartesian Z [km]

    Notes
    -------
    The global cartesian coordinate system is defined as:
        - origin: center of the Earth
        - Z axis in the direction of the rotational axis and through the North
          pole
        - X axis in the equatorial plane and through the prime meridian.
    The local cartesian coordinate system is defined as:
        - origin: local position
        - X: East
        - Y: North
        - Z: up
    The meaning of the input (X,Y,Z) depends on the direction of the conversion
    (to global cartesian or to global spherical).
    """
    # First get global cartesian coordinates of local origin
    (goX, goY, goZ) = globalspherical2globalcartesian(lat, lon, rho)

    if not inverse:
        # Translate global position to local origin
        tx = X - goX
        ty = Y - goY
        tz = Z - goZ
        # Then, rotate about global-Z to get local-X pointing eastward
        rot = -np.radians(lon + 90.0)
        sx = tx * np.cos(rot) - ty * np.sin(rot)
        sy = tx * np.sin(rot) + ty * np.cos(rot)
        sz = tz
        # Finally, rotate about X axis to align Z with upward direction
        rot = -np.radians(90.0 - lat)
        xOut = sx
        yOut = sy * np.cos(rot) - sz * np.sin(rot)
        zOut = sy * np.sin(rot) + sz * np.cos(rot)
    else:
        # First rotate about X axis to align Z with Earth rotational axis
        # direction
        rot = np.radians(90.0 - lat)
        sx = X
        sy = Y * np.cos(rot) - Z * np.sin(rot)
        sz = Y * np.sin(rot) + Z * np.cos(rot)
        # Rotate about global-Z to get global-X pointing to the prime meridian
        rot = np.radians(lon + 90.0)
        xOut = sx * np.cos(rot) - sy * np.sin(rot)
        yOut = sx * np.sin(rot) + sy * np.cos(rot)
        zOut = sz
        # Finally, translate local position to global origin
        xOut = xOut + goX
        yOut = yOut + goY
        zOut = zOut + goZ

    return xOut, yOut, zOut


def localspherical2localcartesian(X, Y, Z, inverse=False):
    """Convert a position from local spherical to local cartesian,
        or vice-versa

    Parameters
    ----------
    X : float
        azimuth [degree, N] or local cartesian X [km]
    Y : float
        elevation [degree] or local cartesian Y [km]
    Z : float
        distance origin [km] or local cartesian Z [km]
    inverse : Optional[bool]
        inverse conversion

    Returns
    -------
    X : float
        local cartesian X [km] or azimuth [degree, N]
    Y : float
        local cartesian Y [km] or elevation [degree]
    Z : float
        local cartesian Z [km] or distance from origin [km]

    Notes
    ------
    The local spherical coordinate system is defined as:
        - origin: local position
        - azimuth (with respect to North)
        - Elevation (with respect to horizon)
        - Altitude
    The local cartesian coordinate system is defined as:
        - origin: local position
        - X: East
        - Y: North
        - Z: up
    The meaning of the input (X,Y,Z) depends on the direction of the conversion
    (to global cartesian or to global spherical).
    """
    if not inverse:
        # local spherical into local cartesian
        r = Z
        el = Y
        az = X
        xOut = r * np.cos(np.radians(el)) * np.sin(np.radians(az))
        yOut = r * np.cos(np.radians(el)) * np.cos(np.radians(az))
        zOut = r * np.sin(np.radians(el))
    else:
        # local cartesian into local spherical
        r = np.sqrt(X**2 + Y**2 + Z**2)
        el = np.degrees(np.arcsin(Z / r))
        az = np.degrees(np.arctan2(X, Y))
        xOut = az
        yOut = el
        zOut = r

    return xOut, yOut, zOut


def geodetic2geocAzEl(lat, lon, az, el, inverse=False):
    """Converts pointing azimuth and elevation measured with respect to the
    local horizon to azimuth and elevation with respect to the horizon defined
    by the plane perpendicular to the Earth-centered radial vector drawn
    through a user defined point.

    Parameters
    ----------
    lat : float
        latitude [degree]
    lon : float
        longitude [degree]
    az : float
        azimuth [degree, N]
    el : float
        elevation [degree]
    inverse : Optional[bool]
        inverse conversion

    Returns
    -------
    lat : float
        latitude [degree]
    lon : float
        longitude [degree]
    Re : float
        Earth radius [km]
    az : float
        azimuth [degree, N]
    el : float
        elevation [degree]
    """
    taz = np.radians(az)
    tel = np.radians(el)

    # In this transformation x is east, y is north and z is up
    if not inverse:
        # Calculate deviation from vertical (in radians)
        (geocLat, geocLon, Re) = geod2geoc(lat, lon)
        devH = np.radians(lat - geocLat)
        # Calculate cartesian coordinated in local system
        kxGD = np.cos(tel) * np.sin(taz)
        kyGD = np.cos(tel) * np.cos(taz)
        kzGD = np.sin(tel)
        # Now rotate system about the x axis to align local vertical vector
        # with Earth radial vector
        kxGC = kxGD
        kyGC = kyGD * np.cos(devH) + kzGD * np.sin(devH)
        kzGC = -kyGD * np.sin(devH) + kzGD * np.cos(devH)
        # Finally calculate the new azimuth and elevation in the geocentric
        # frame
        azOut = np.degrees(np.arctan2(kxGC, kyGC))
        elOut = np.degrees(np.arctan(kzGC / np.sqrt(kxGC**2 + kyGC**2)))
        latOut = geocLat
        lonOut = geocLon
    else:
        # Calculate deviation from vertical (in radians)
        (geodLat, geodLon, Re) = geod2geoc(lat, lon, inverse=True)
        devH = np.radians(geodLat - lat)
        # Calculate cartesian coordinated in geocentric system
        kxGC = np.cos(tel) * np.sin(taz)
        kyGC = np.cos(tel) * np.cos(taz)
        kzGC = np.sin(tel)
        # Now rotate system about the x axis to align local vertical vector
        # with Earth radial vector
        kxGD = kxGC
        kyGD = kyGC * np.cos(-devH) + kzGC * np.sin(-devH)
        kzGD = -kyGC * np.sin(-devH) + kzGC * np.cos(-devH)
        # Finally calculate the new azimuth and elevation in the geocentric
        # frame
        azOut = np.degrees(np.arctan2(kxGD, kyGD))
        elOut = np.degrees(np.arctan(kzGD / np.sqrt(kxGD**2 + kyGD**2)))
        latOut = geodLat
        lonOut = geodLon

    return latOut, lonOut, Re, azOut, elOut


def calculate_azimuth(slat, slon, salt, elat, elon, ealt):
    """
    calculates the azimuth between two given points
    Only one method used and only azimuth returned, this can be expanded
    using XS and SC existing codebase

    Parameters
    ----------
    slat : float
        origin latitude [degree]
    slon : float
        origin longitude [degree]
    salt : float
        origin altitude [km]
    elat : float
        distant latitude [degree]
    elon : float
        distant longitude [degree]
    ealt : float
        distant altitude [km]

    Returns
    -------
    azm : float
        azimuth between origin location and distant location [degrees]
        in reference to North
    """
    # Convert point of origin from geodetic to geocentric
    gclat, gclon, srho = geod2geoc(slat, slon)
    # Convert distant point from geodetic to geocentric
    gcelat, gcelon, erho = geod2geoc(elat, elon)
    # Convert distant point from geocentric to global cartesian
    (pX, pY, pZ) = globalspherical2globalcartesian(gcelat, gcelon, erho + ealt)
    # Convert pointing direction from global cartesian to local cartesian
    (dX, dY, dZ) = globalcartesian2localcartesian(pX, pY, pZ, gclat,
                                                  gclon, srho + salt)
    # Convert pointing direction from local cartesian to local spherical
    (gaz, gel, rho) = localspherical2localcartesian(dX, dY, dZ, inverse=True)
    # Convert pointing azimuth and elevation to geodetic
    (_, _, _, azm, _) = geodetic2geocAzEl(gclat, gclon, gaz, gel, inverse=True)
    return azm



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Daniel Billett, Marina Schmidt
#
# Modifications:
#   20220308 MTS added partial record exception
#   20230628 CJM refactored return values
#   20230713 CJM corrected geographic quivers
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.


"""
Grid plots
"""

import datetime as dt
import matplotlib.pyplot as plt
import numpy as np
import warnings

from matplotlib import ticker, cm, colors
from typing import List

# Third party libraries
import aacgmv2

from pydarn import (PyDARNColormaps, Fan, plot_exceptions, Hemisphere, RadarID,
                    standard_warning_format, Projs, Coords, GeneralUtils)

warnings.formatwarning = standard_warning_format


class Grid():
    """
        Grid plots for SuperDARN data
        This class inherits from matplotlib to generate the figures
        Methods
        -------
        plot_grid
    """
    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_grid()\n"

    @classmethod
    def plot_grid(cls, dmap_data: List[dict], record: int = 0,
                  start_time: dt.datetime = None, time_delta: int = 1,
                  ax=None, parameter: str = 'vel',
                  cmap: str = None, zmin: int = None,
                  zmax: int = None, colorbar: bool = True,
                  colorbar_label: str = '', title: str = '',
                  len_factor: float = 150.0, ref_vector: int = 300,
                  projs: Projs = Projs.POLAR,
                  coords: Coords = Coords.AACGM_MLT,
                  **kwargs):
        """
        Plots a radar's gridded vectors from a GRID file

        Parameters
        -----------
            dmap_data: List[dict]
                Named list of dictionaries obtained from SDarn_read
            record: int
                record number to plot
                default: 0
            start_time: datetime.datetime
                datetime object as the start time of the record to plot
                if none then record will be used
                default: none
            time_delta: int
                How close the start_time has to be start_time of the record
                in minutes
                default: 1
            ax: matplotlib.pyplot axis
                Pre-defined axis object to pass in, must currently
                be polar projection
                Default: Generates a polar projection for the user
                with MLT/latitude labels
            parameter: str
                Key name indicating which parameter to plot.
                Default: vel (Velocity). Alternatives: 'pwr', 'wdt'
            cmap: matplotlib.cm
                matplotlib colour map
                https://matplotlib.org/tutorials/colors/colormaps.html
                Default: Official pyDARN colour map for given parameter
            zmin: int
                The minimum parameter value for coloring
                Default: {'pwr': [0], 'vel': [0], 'wdt': [0]}
            zmax: int
                The maximum parameter value for  coloring
                Default: {'pwr': [50], 'vel': [1000], 'wdt': [250]}
            colorbar: bool
                Draw a colourbar if True
                Default: True
            colorbar_label: str
                The label that appears next to the colour bar.
                Requires colorbar to be true
                Default: ''
            title: str
                Adds a title to the plot. If no title is specified,
                one will be provided
                Default: ''
            len_factor: float
                Normalisation factor for the vectors, to control size on plot
                Larger number means smaller vectors on plot
                Default: 150.0
            ref_vector: int
                Velocity value to be used for the reference vector, in m/s
                Default: 300
            projs: Enum
                choice of projection for plot
                default: Projs.POLAR (polar projection)
            coords: Enum
                choice of plotting coordinates
                default: Coords.AACGM_MLT (Magnetic Lat and MLT)
            kwargs: key=value
                uses the parameters for plot_fov and projections.axis
        See Also
        --------
        plot_fov - plots the field of view found in fan.py

        Returns
        -----------
        If parameter is 'vel':
        thetas - List of gridded data point magnetic local times (degrees)
        end_thetas - List of magnetic local time end points used for vector
                     plotting (degrees)
        rs - List of gridded data point radius' (latitude)
        end_rs - List of radius end points for vector plotting (latitude)
        data - List of magnitudes of line-of-sight velocity
        azm_v -  List of azimuths of line-of-sight velocity
        else:
        thetas - List of gridded data point magnetic local times (degrees)
        rs - List of gridded data point radius' (latitude)
        data - List of data magnitudes plotted, for parameter chosen
        """
        # Short hand for the parameters in GRID files
        if parameter == 'vel' or parameter == 'pwr' or parameter == 'wdt':
            parameter = "vector.{param}.median".format(param=parameter)

        # Find the record corresponding to the start time
        if start_time is not None:
            for record in range(len(dmap_data)):
                date = dt.datetime(dmap_data[record]['start.year'],
                                   dmap_data[record]['start.month'],
                                   dmap_data[record]['start.day'],
                                   dmap_data[record]['start.hour'],
                                   dmap_data[record]['start.minute'])
                time_diff = date - start_time
                if time_diff.seconds/60 <= time_delta:
                    break
            if time_diff.seconds/60 > time_delta:
                raise plot_exceptions.NoDataFoundError(parameter,
                                                       start_time=start_time)
        else:
            # Record is read in or default to 0
            date = dt.datetime(dmap_data[record]['start.year'],
                               dmap_data[record]['start.month'],
                               dmap_data[record]['start.day'],
                               dmap_data[record]['start.hour'],
                               dmap_data[record]['start.minute'])

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # Check for partial records
            if all(dmap_data[record]['nvec'] == 0):
                raise plot_exceptions.PartialRecordsError('~all vectors~')

            data_lons = dmap_data[record]['vector.mlon']
            data_lats = dmap_data[record]['vector.mlat']

            # Hemisphere is not found in grid files so take from latitudes
            hemisphere = Hemisphere(np.sign(
                                    dmap_data[record]['vector.mlat'][0]))
            ax, ccrs = projs(date=date, ax=ax, hemisphere=hemisphere, **kwargs)
            if ccrs is None:
                transform = ax.transData
            else:
                transform = ccrs.Geodetic()

            for stid in dmap_data[record]['stid']:
                fan_rtn = Fan.plot_fov(RadarID(stid), date, ax=ax, ccrs=ccrs,
                                       coords=coords, projs=projs, **kwargs)
                coord_lons = fan_rtn['data']['beam_corners_lons']
                ax = fan_rtn['ax']
                ccrs = fan_rtn['ccrs']

            if coords != Coords.GEOGRAPHIC and projs == Projs.GEO:
                raise plot_exceptions.NotImplemented(
                                        "AACGM coordinates are"
                                        " not implemented for "
                                        " geographic projections right"
                                        " now, if you would like to"
                                        " see it sooner please help"
                                        " out at "
                                        "https://github.com"
                                        "/SuperDARN/pyDARN")
            if coords == Coords.GEOGRAPHIC and projs == Projs.POLAR:
                raise plot_exceptions.NotImplemented(
                                        "Geographic coordinates are"
                                        " not implemented for "
                                        " polar projections in grid plots"
                                        " right now, if you would like to"
                                        " see it sooner please help"
                                        " out at "
                                        "https://github.com"
                                        "/SuperDARN/pyDARN")
            # Thetas in aacgm required for calculating the
            # vectors later, called them theta_calc and rs_calc
            # for later use
            shifted_mlts = coord_lons[0, 0] - \
                (aacgmv2.convert_mlt(coord_lons[0, 0], date) * 15)
            shifted_lons = data_lons - shifted_mlts

            if projs == Projs.POLAR:
                # Convert to radians for polar plots
                thetas_calc = np.radians(shifted_lons)
                thetas = np.radians(shifted_lons)
                rs = data_lats
                rs_calc = data_lats
            elif projs == Projs.MAG:
                thetas_calc = np.radians(shifted_lons)
                thetas = shifted_lons
                rs = data_lats
                rs_calc = data_lats
            else:
                # Convert to geographic coordinates
                glats, glons, _ = aacgmv2.convert_latlon_arr(
                                    data_lats, data_lons, 300,
                                    date, method_code="A2G")
                thetas_calc = np.radians(glons)
                thetas = glons
                rs = glats
                rs_calc = glats

            # Colour table and max value selection depending on
            # parameter plotted Load defaults if none given
            if cmap is None:
                cmap = {'vector.pwr.median': PyDARNColormaps.PYDARN_PLASMA,
                        'vector.vel.median': PyDARNColormaps.PYDARN_PLASMA_R,
                        'vector.wdt.median': PyDARNColormaps.PYDARN_VIRIDIS}
                cmap = cmap[parameter]

            # Setting zmin and zmax
            defaultzminmax = {'vector.pwr.median': [0, 50],
                              'vector.vel.median': [0, 1000],
                              'vector.wdt.median': [0, 250]}
            if zmin is None:
                zmin = defaultzminmax[parameter][0]
            if zmax is None:
                zmax = defaultzminmax[parameter][1]

            norm = colors.Normalize
            norm = norm(zmin, zmax)

            # check to make sure the parameter is present in the file
            # this may not be the case for wdt and pwr as you need -xtd
            # option in make_grid
            try:
                data = dmap_data[record][parameter]
            except KeyError:
                raise plot_exceptions.UnknownParameterError(parameter,
                                                            grid=True)
            # Plot the magnitude of the parameter
            ax.scatter(thetas, rs, c=data, s=2.0, vmin=zmin, vmax=zmax,
                       zorder=5, cmap=cmap, transform=transform)

            # If the parameter is velocity then plot the LOS vectors
            if parameter == "vector.vel.median":

                # Get the mag azimuths from the data
                azm_v = dmap_data[record]['vector.kvect']

                # Azimuth direction needs to be converted to geographic
                if projs == Projs.GEO:
                    # Pole coords in geo for this date
                    gpole_lat = 90 * hemisphere.value
                    gpole_lon = 0
                    mpole_lat, mpole_lon, _ = \
                        aacgmv2.convert_latlon(gpole_lat, gpole_lon, 300,
                                               date, method_code='A2G')

                    # Flatten array to work on
                    a_m_flat = []
                    b_m_flat = []
                    thetas_flat = thetas.flatten()
                    rs_flat = rs.flatten()
                    for i in range(len(rs_flat)):
                        a_m_flat.append(
                            GeneralUtils.great_circle(thetas_flat[i],
                                                      rs_flat[i],
                                                      gpole_lon,
                                                      gpole_lat))
                        b_m_flat.append(
                            GeneralUtils.great_circle(thetas_flat[i],
                                                      rs_flat[i],
                                                      mpole_lon,
                                                      mpole_lat))
                    a_m = np.reshape(a_m_flat, rs.shape)
                    b_m = np.reshape(b_m_flat, rs.shape)
                    c_m_flat = np.ones(len(rs_flat)) * \
                        GeneralUtils.great_circle(gpole_lon, gpole_lat,
                                                  mpole_lon, mpole_lat)
                    c_m = np.reshape(c_m_flat, rs.shape)
                    declination = np.arccos((np.cos(c_m) - np.cos(a_m)
                                             * np.cos(b_m)) /
                                            (np.sin(a_m) * np.sin(b_m)))

                    azm_v = azm_v + np.degrees(declination) * hemisphere.value

                # Number of data points
                num_pts = range(len(data))

                # Angle to "rotate" each vector by to get into same
                # reference frame Controlled by longitude, or "mltitude"
                alpha = thetas_calc

                # Convert initial positions to Cartesian
                start_pos_x = (90 - abs(rs_calc)) * np.cos(thetas_calc)
                start_pos_y = (90 - abs(rs_calc)) * np.sin(thetas_calc)

                # Resolve LOS vector in x and y directions,
                # with respect to mag pole
                # Gives zonal and meridional components of LOS vector
                los_x = -data * np.cos(np.radians(
                                       -azm_v * hemisphere.value))
                los_y = -data * np.sin(np.radians(
                                       -azm_v * hemisphere.value))

                # Rotate each vector into same reference frame
                # following vector rotation matrix
                # https://en.wikipedia.org/wiki/Rotation_matrix
                vec_x = (los_x * np.cos(alpha)) - (los_y * np.sin(alpha))
                vec_y = (los_x * np.sin(alpha)) + (los_y * np.cos(alpha))

                # New vector end points, in Cartesian
                end_pos_x = start_pos_x + (vec_x * hemisphere.value /
                                           len_factor)
                end_pos_y = start_pos_y + (vec_y * hemisphere.value /
                                           len_factor)

                # Convert back to polar for plotting
                end_rs = 90 - (np.sqrt(end_pos_x**2 + end_pos_y**2))
                end_thetas = np.arctan2(end_pos_y, end_pos_x)

                end_rs = end_rs * hemisphere.value

                # Plot the vectors
                if projs == Projs.POLAR:
                    for i in num_pts:
                        ax.plot([thetas[i], end_thetas[i]],
                                 [rs[i], end_rs[i]], c=cmap(norm(data[i])),
                                 linewidth=0.5, transform=transform)
                else:
                    # If proj is geographic
                    end_thetas = np.degrees(end_thetas)
                    for i in num_pts:
                        # If the vector does not cross the meridian
                        if np.sign(thetas[i]) == np.sign(end_thetas[i]):
                            ax.plot([thetas[i], end_thetas[i]],
                                     [rs[i], end_rs[i]],
                                     c=cmap(norm(data[i])),
                                     linewidth=0.5, transform=transform)
                        # If the vector crosses the meridian then amend so that
                        # the start and end are in the same sign
                        else:
                            if abs(end_thetas[i]) > 90:
                                if end_thetas[i] < 0:
                                    end_thetas[i] = end_thetas[i] + 360
                                else:
                                    thetas[i] = thetas[i] + 360
                            # Vector plots correctly over the 0 meridian so
                            # Nothing is done to correct that section
                            ax.plot([thetas[i], end_thetas[i]],
                                     [rs[i], end_rs[i]],
                                     c=cmap(norm(data[i])),
                                     linewidth=0.5, transform=transform)
                # TODO: Add a velocity reference vector

        if colorbar is True:
            mappable = cm.ScalarMappable(norm=norm, cmap=cmap)
            locator = ticker.MaxNLocator(symmetric=True, min_n_ticks=3,
                                         integer=True, nbins='auto')
            ticks = locator.tick_values(vmin=zmin, vmax=zmax)

            if zmin == 0:
                cb = ax.figure.colorbar(mappable, ax=ax, extend='max',
                                        ticks=ticks)
            else:
                cb = ax.figure.colorbar(mappable, ax=ax, extend='both',
                                        ticks=ticks)

            if colorbar_label != '':
                cb.set_label(colorbar_label)
        else:
            cb = None

        if title == '':
            title = "{year}-{month}-{day} {start_hour}:{start_minute} -"\
                " {end_hour}:{end_minute}"\
                    "".format(year=date.year,
                              month=str(date.month).zfill(2),
                              day=str(date.day).zfill(2),
                              start_hour=str(date.hour).zfill(2),
                              start_minute=str(date.minute).zfill(2),
                              end_hour=str(dmap_data[record]['end.hour']).
                              zfill(2),
                              end_minute=str(dmap_data[record]['end.minute']).
                              zfill(2))
        plt.title(title)
        if parameter != 'vector.vel.median':
            end_thetas = None
            end_rs = None
            azm_v = None
        return {'ax': ax,
                'ccrs': ccrs,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'data_position_theta': thetas,
                         'end_stick_position_theta': end_thetas,
                         'data_position_r': rs,
                         'end_stick_position_r': end_rs,
                         'raw_data': data,
                         'raw_azimuths': azm_v}
                }



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Authors: Marina Schmidt and Danno Peters
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modifications:
# 2022-02-11 MTS USASK updated the _HDW_info class to take in
#            the hardware format
# 2023-01-21 CJM Added ICE and ICW defaults and hdw link
# 2024-01-24 CJM added NSSC radars and updated hdw link
"""
This module contains SuperDARN radar information
"""
import glob
import os
import pydarn
import shutil

from typing import NamedTuple
from enum import Enum
from datetime import datetime
from subprocess import check_call


def get_hdw_files(force: bool = True, version: str = None):
    """
    downloads hardware files from the SuperDARN github page:
        https://github.com/SuperDARN/hdw

    Parameter
    ---------
    force: bool
        download hardware files even if they are in the
        directory
    version: str
        version number to download

    Note: version is not currently working as hardware files
    have yet to be versioned.
    """

    # Path should the path where pydarn is installed
    hdw_path = "{}/hdw/".format(os.path.dirname(pydarn.utils.__file__))

    # TODO: implement when DSWG starts versioning hardware files
    if version is not None:
        raise Exception("This feature is not implemented yet")

    # if there is no files in hdw folder or force is true
    # download the hdw files
    if len(os.listdir(hdw_path)) == 0 or force:
        # pycurl doesn't download a zip folder easily so
        # use the command line command
        check_call(['curl', '-L', '-o', hdw_path+'/main.zip',
                    'https://github.com/SuperDARN/hdw/archive/main.zip'])
        # use unzip command because zipfile on works with files and not folders
        # though this is possible with zipfile but this was easier for me to
        # get it working
        check_call(['unzip', '-d', hdw_path, hdw_path+'/main.zip'])
        dat_files = glob.glob(hdw_path+'/hdw-main/*')
        # shutil only moves specific files so we need to move
        # everything one at a time
        for hdw_file in dat_files:
            shutil.move(hdw_file, hdw_path+os.path.basename(hdw_file))
        # delete the empty folder
        os.removedirs(hdw_path+'/hdw-main/')


def read_hdw_file(abbrv, date: datetime = None, update: bool = False):
    """
    Reads the hardware file for the associated abbreviation of the radar name.

    Parameters
    ----------
        abbrv : str
            Radars 3 letter assigned abbreviation
        date: datetime
            Datetime object of hardware information to obtain
            default: current date
        update: bool
            If True this will update the hardware files again
            without re-installing pydarn
            default: False
    Return
    ------
    _HdwInfo object that contains all the field names in a hardware
        file

    Raises
    ------
    HardwareFileNotFoundError raised when there is no hardware file found for
    the given abbreviation
    """
    if date is None:
        date = datetime.now()

    hdw_path = os.path.dirname(__file__)+'/hdw/'
    hdw_file = "{path}/hdw.dat.{radar}".format(path=hdw_path, radar=abbrv)
    hdw_data = []
    hdw_lines_date = []
    # if the file does not exist then try
    # and download it
    if os.path.exists(hdw_file) is False:
        get_hdw_files(force=update)
    try:
        with open(hdw_file, 'r') as reader:
            lines = reader.readlines()
            for i in range(len(lines)):
                if '#' not in lines[i] and len(lines[i].split()) > 1:
                    hdw_data.append(lines[i].split())
                    """
                    Hardware files give the year and seconds from the beginning
                    of that year. Thus to check the date if it corresponds we
                    need to convert to a datetime object and then compare.
                    """
                    j = len(hdw_data)-1
                    hdw_lines_date.append(
                        datetime(year=int(hdw_data[j][2][0:4]),
                                 month=int(hdw_data[j][2][4:6]),
                                 day=int(hdw_data[j][2][6:8]),
                                 hour=int(hdw_data[j][3][0:2]),
                                 minute=int(hdw_data[j][3][3:5]),
                                 second=int(hdw_data[j][3][6:8])))
                    if hdw_lines_date[j] > date:
                        j = j-1
                        break
            """
            Hardware data array positions definitions:
                0: Station ID (unique numerical value).
                1: Status code (1 operational, -1 offline).
                2: First date that parameter string is valid
                   (YYYYMMDD).
                3: First time that parameter string is valid
                   (HH:MM:SS).
                4: Geographic latitude of radar site
                   (Given in decimal degrees to 3
                   decimal places. Southern hemisphere
                   values are negative)
                5: Geographic longitude of radar site
                   (Given in decimal degrees to
                   3 decimal places.
                   West longitude values are negative)
                6: Altitude of the radar site (meters)
                7: Physical scanning boresight
                   (Direction of the center beam, measured in
                   degrees relative to geographic north.
                   CCW rotations are negative.)
                8: Electronic shift to radar scanning
                   boresight (Degrees relative to
                   physical antenna boresight.
                   Normally 0.0 degrees)
                9: Beam separation (Angular
                   separation in degrees between adjacent
                   beams. Normally 3.24 degrees)
                10: Velocity sign (At the radar level,
                    backscattered signals with
                    frequencies above the transmitted
                    frequency are assigned positive
                    Doppler velocities while backscattered
                    signals with frequencies below
                    the transmitted frequency are assigned
                    negative Doppler velocity. This
                    convention can be reversed by changes
                    in receiver design or in the
                    data sampling rate. This parameter
                    is set to +1 or -1 to maintain the
                    convention.)
                11: Phase sign (Cabling errors can
                    lead to a 180 degree shift of the
                    interferometry phase measurement.
                    +1 indicates that the sign is
                    correct, -1 indicates that it must be flipped.)
                12: Tdiff [Channel A]
                    (Propagation time from interferometer
                    array antenna to phasing matrix input
                    minus propagation time from main array antenna
                    through transmitter to phasing matrix input.
                    Units are decimal
                    microseconds)
                13: Tdiff [Channel B]
                    (Propagation time from interferometer
                    array antenna to phasing matrix input minus
                    propagation time from main array antenna
                    through transmitter to phasing matrix input.
                    Units are decimal microseconds)
                14: Interferometer X offset
                    (Displacement of midpoint of interferometer
                    array from midpoint of main array,
                    along the line of antennas
                    with +X toward higher antenna numbers.
                    Units are meters)
                15: Interferometer Y offset
                    (Displacement of midpoint of
                    interferometer array from midpoint of
                    main array, along the array
                    normal direction with +Y in the direction of
                    the array normal. Units are meters)
                16: Interferometer Z offset
                    (Displacement of midpoint of
                    interferometer array from midpoint of
                    main array, in terms of altitude
                    difference with +Z up. Units are meters)
                17: Analog Rx rise time
                    (Time given in microseconds. Time delays of
                    less than ~10 microseconds can be ignored.
                    If narrow-band filters are
                    used in analog receivers or front-ends,
                    the time delays should be
                    specified.)
                18: Analog Rx attenuator step (dB)
                19: Analog attenuation stages (Number of stages.
                    This is used for gain control of an analog
                    receiver or front-end.)
                20: Maximum of range gates used
                21: Maximum number of beams
            """
            return _HdwInfo(stid=int(hdw_data[j][0]),
                            status=Status(int(hdw_data[j][1])),
                            abbrev=abbrv,
                            date=hdw_lines_date[j],
                            geographic=_Coord(float(hdw_data[j][4]),
                                              float(hdw_data[j][5]),
                                              float(hdw_data[j][6])),
                            boresight=_Boresight(float(hdw_data[j][7]),
                                                 float(hdw_data[j][8])),
                            beam_separation=float(hdw_data[j][9]),
                            velocity_sign=float(hdw_data[j][10]),
                            phase_sign=float(hdw_data[j][11]),
                            tdiff=_Tdiff(float(hdw_data[j][12]),
                                         float(hdw_data[j][13])),
                            interferometer_offset=_InterferometerOffset(
                                float(hdw_data[j][14]),
                                float(hdw_data[j][15]),
                                float(hdw_data[j][16])),
                            rx_rise_time=float(hdw_data[j][17]),
                            rx_attenuator=float(hdw_data[j][18]),
                            attenuation_stages=int(hdw_data[j][19]),
                            gates=int(hdw_data[j][20]),
                            beams=int(hdw_data[j][21]))
    except FileNotFoundError:
        raise pydarn.radar_exceptions.HardwareFileNotFoundError(abbrv)


class Hemisphere(Enum):
    """
    Class used to denote which hemisphere a radar is located in

    Attributes
    ----------
    Hemisphere.North = 1
    Hemisphere.South = -1

    Notes
    -----
    This is based on the values assigned to the MAP file hemisphere
    field in RST>
    """
    North = 1
    South = -1


class Status(Enum):
    """
    this class is to give names to status of radars

    Attributes
    ----------
    online : 1
    offline : -1
    """
    online = 1
    offline = -1
    other = 0


class _InterferometerOffset(NamedTuple):
    """
    Named tuple class to contain the interferometer offset
    Cartesian coordinates.

    Attributes
    ----------
    x : float
        direction along the line of antennas with +X toward
        higher antenna number
    y : float
        direction along the array normal with +Y in the
        direction of the array normal
    z : float
        direction is the altitude difference, +Z up
    """
    x: float
    y: float
    z: float


class _Coord(NamedTuple):
    """
    Class used to store Geo Coordinates (utilized for both Geographic
    and Geomagnetic) representing location and Boresite direction

    Attributes
    ----------
    lat : float
        Latitude in decimal format
    lon : float
        Longitude in decimal format
    alt : float
        Altitude in meters
    """
    lat: float
    lon: float
    alt: float


class _Tdiff(NamedTuple):
    """
    Class use to store tdiff values for various channels and
    maybe for common frequencies.

    Attributes
    ----------
    channel_a : float
        for channel a
    channel_b : float
        for channel b
    """
    channel_a: float
    channel_b: float


class _Boresight(NamedTuple):
    """
    This class used to store the various boresights

    Attributes
    ----------
    physical : float
       Physical scanning boresight
    electronic : float
       Electronic shift to the radar scanning boresight
    """
    physical: float
    electronic: float


class _HdwInfo(NamedTuple):
    """
    Class used to store relevant information about the SuperDARN Radars

    Attributes
    --- -------
    stid : int
        station number
    status : Status
        status of radars operation
    abbrev : str
        three letter station abbreviation
    date : datetime
        date of the hdw file data is being used
    geographic : _Coord object
        Named Tuple containing geographic latitude longitude and altitude
    boresight : _Boresight
        boresight physical and electronic in degrees
    beam_separation : float
        angular separation between radar beams
    velocity_sign : float
        at radar level, backscattered signal with
        frequencies above the transmitted frequency are assigned positive
        Doppler velocities while backscattered signals with frequencies
        below the transmitted frequency are assigned negative Doppler
        velocity. Can be changed in receiver design.
    phase_sign : float
        To account for flipped cable errors
    tdiff : _Tdiff
        channel a and b:
        propagation time from interferometer array antenna
        to phasing matrix input minus propagation time from main array
        antenna through transmitter to phasing matrix input.
        (microseconds)
    interferometer_offset : _InterferometerOffset
        displacement of midpoint
        interferometer array from midpoint main array in
        Cartesian coordinates(meters).
    rx_attenuator : float
        Analog Rx attenuator step (dB)
    rx_rise : float
        Analog Rx rise time (microseconds)
    attenuation_stages : float
        Analog Attenuation stages - gain control of an analog
        receiver or front-end
    gates :int
        number of range gates per beam
    beams : int
        number of possible beams

    See Also
    --------
    read_hdw_file : function for reading hardware files
    Status
    _Coord : object contain coordinate information
    _Boresight
    _Tdiff
    _Interferometer
    """
    stid: int
    status: int
    abbrev: str
    date: datetime
    geographic: _Coord
    boresight: _Boresight
    beam_separation: float
    velocity_sign: float
    rx_attenuator: float
    tdiff: _Tdiff
    phase_sign: float
    interferometer_offset: _InterferometerOffset
    rx_rise_time: float
    attenuation_stages: float
    gates: int
    beams: int


class _Radar(NamedTuple):
    """
    Class used to combine all Radar information
    Attributes
    ----------
    name: str
        Name of the Radar
    institute: str
        Institution of the Radar
    hemisphere: Hemisphere
        Hemisphere the radar belongs to
    hardware_info: _HdwInfo
        NamedTuple containing all hardware information from hardware files
    """
    name: str
    institution: str
    hemisphere: Hemisphere
    range_gate_45: int
    geo_label: list
    mag_label: list
    hardware_info: _HdwInfo


class RadarID(Enum):
    """
    Class used to denote the Station ID (stid) of each radar.

    Attributes
    ----------
    Three-letter site code = station ID
    for each site
    """
    # Northern hemisphere radars
    ADE = 209
    ADW = 208
    BKS = 33
    CVE = 207
    CVW = 206
    CLY = 66
    FHE = 205
    FHW = 204
    GBR = 1
    HAN = 10
    HJE = 55
    HJW = 56
    HOK = 40
    HKW = 41
    ICE = 211
    ICW = 210
    INV = 64
    JME = 50
    KAP = 3
    KSR = 16
    KOD = 7
    LJE = 51
    LJW = 52
    LYR = 90
    PYK = 9
    PGR = 6
    RKN = 65
    SAS = 5
    SCH = 2
    SZE = 53
    SZW = 54
    STO = 8
    WAL = 32
    # Southern hemisphere radars
    BPK = 24
    DCE = 96
    DCN = 97
    FIR = 21
    HAL = 4
    KER = 15
    MCM = 20
    SAN = 11
    SPS = 22
    SYE = 13
    SYS = 12
    TIG = 14
    UNW = 18
    ZHO = 19


class SuperDARNRadars():
    """
    Class containing a dictionary of Nested Named Tuples with information
    about each radar in SuperDARN.

    Attributes
    ----------
        radars: dict
            dictionary of each SuperDARN radar with key being STID value and
            a _Radar object containing the name, institutional and hardware
            information of the radar

    See Also
    --------
        _Radar : radar object containing radar information
        read_hdw_file : function to read hardware information for a given radar
    """
    radars = {RadarID.ADE: _Radar('Adak Island East',
                          'University of Alaska Fairbanks', Hemisphere.North,
                                  75, [47, -172], [42, -106], read_hdw_file('ade')),
              RadarID.ADW: _Radar('Adak Island West', 'University of Alaska Fairbanks',
                                  Hemisphere.North, 75, [47, 178], [42, -116],
                                  read_hdw_file('adw')),
              RadarID.BKS: _Radar('Blackstone', 'Virginia Tech', Hemisphere.North,
                                  100, [32, -78], [44, -5], read_hdw_file('bks')),
              RadarID.CVE: _Radar('Christmas Valley East', 'Dartmouth College',
                                  Hemisphere.North, 100, [38, -115], [48, -53],
                                  read_hdw_file('cve')),
              RadarID.CVW: _Radar('Christmas Valley West', 'Dartmouth College',
                                  Hemisphere.North, 100, [38, -125], [48, -63],
                                  read_hdw_file('cvw')),
              RadarID.CLY: _Radar('Clyde River', 'University of Saskatchewan',
                                  Hemisphere.North, 100, [65, -68], [72, 17],
                                  read_hdw_file('cly')),
              RadarID.FHE: _Radar('Fort Hays East', 'Virginia Tech', Hemisphere.North,
                                  100, [34, -94], [45, -25], read_hdw_file('fhe')),
              RadarID.FHW: _Radar('Fort Hays West', 'Virginia Tech', Hemisphere.North,
                                  100, [34, -104], [45, -35], read_hdw_file('fhw')),
              RadarID.GBR: _Radar('Goose Bay', 'Virginia Tech', Hemisphere.North,
                                  100, [48, -60], [54, 23], read_hdw_file('gbr')),
              RadarID.HAN: _Radar('Hankasalmi', 'University of Leicester',
                                  Hemisphere.North, 70, [57, 27], [54, 102],
                                  read_hdw_file('han')),
              RadarID.HJE: _Radar('Hejing East',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [40, 87], [36, 163],
                                  read_hdw_file('hje')),
              RadarID.HJW: _Radar('Hejing West',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [40, 81], [36, 153],
                                  read_hdw_file('hjw')),
              RadarID.HOK: _Radar('Hokkaido East', 'Nagoya University',
                                  Hemisphere.North, 110, [39, 149], [35, -139],
                                  read_hdw_file('hok')),
              RadarID.HKW: _Radar('Hokkaido West', 'Nagoya University',
                                  Hemisphere.North, 110, [39, 139], [35, -149],
                                  read_hdw_file('hkw')),
              RadarID.ICE: _Radar('Iceland East', 'Dartmouth College',
                                  Hemisphere.North, 100, [61, -16], [60, 70],
                                  read_hdw_file('ice')),
              RadarID.ICW: _Radar('Iceland West', 'Dartmouth College',
                                  Hemisphere.North, 100, [61, -26], [60, 60],
                                  read_hdw_file('icw')),
              RadarID.INV: _Radar('Inuvik', 'University of Saskatchewan',
                                  Hemisphere.North, 75, [63, -134], [66, -80],
                                  read_hdw_file('inv')),
              RadarID.JME: _Radar('Jiamusi East',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [42, 130], [37, -155],
                                  read_hdw_file('jme')),
              RadarID.KAP: _Radar('Kapuskasing', 'Virginia Tech', Hemisphere.North,
                                  75, [44, -82], [54, -7], read_hdw_file('kap')),
              RadarID.KSR: _Radar('King Salmon',
                         'National Institute of Information and'
                         ' Communications Technology', Hemisphere.North,
                                  75, [54, -162], [52, -99], read_hdw_file('ksr')),
              RadarID.KOD: _Radar('Kodiak', 'University of Alaska Fairbanks',
                                  Hemisphere.North, 110, [53, -152], [52, -92],
                                  read_hdw_file('kod')),
              RadarID.LJE: _Radar('Longjing East',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [39, 132], [32, -151],
                                  read_hdw_file('lje')),
              RadarID.LJW: _Radar('Longjing West',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [39, 126], [42, -161],
                                  read_hdw_file('ljw')),
              RadarID.LYR: _Radar('Longyearbyen', 'University of Centre in Svalbard',
                                  Hemisphere.North, 70, [73, 16], [71, 108],
                                  read_hdw_file('lyr')),
              RadarID.PYK: _Radar('Pykkvibaer', 'University of Leicester',
                                  Hemisphere.North, 70, [58, -19], [56, 75],
                                  read_hdw_file('pyk')),
              RadarID.PGR: _Radar('Prince George', 'University of Saskatchewan',
                                  Hemisphere.North, 75, [49, -123], [55, -61],
                                  read_hdw_file('pgr')),
              RadarID.RKN: _Radar('Rankin Inlet', 'University of Saskatchewan',
                                  Hemisphere.North, 75, [58, -92], [66, -21],
                                  read_hdw_file('rkn')),
              RadarID.SAS: _Radar('Saskatoon', 'University of Saskatchewan',
                                  Hemisphere.North, 75, [47, -107], [56, -41],
                                  read_hdw_file('sas')),
              RadarID.SCH: _Radar('Schefferville', 'CNRS/LPCE', Hemisphere.North,
                                  75, [50, -67], [60, 14], read_hdw_file('sch')),
              RadarID.SZE: _Radar('Siziwanqi East',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [38, 115], [37, -169],
                                  read_hdw_file('sze')),
              RadarID.SZW: _Radar('Siziwanqi West',
                         'National Space Science Center,'
                         'Chinese Academy of Sciences',
                                  Hemisphere.North, 100, [38, 109], [37, -179],
                                  read_hdw_file('szw')),
              RadarID.STO: _Radar('Stokkseyri', 'Lancaster University',
                                  Hemisphere.North, 75, [58, -29], [56, 65],
                                  read_hdw_file('sto')),
              RadarID.WAL: _Radar('Wallops Island', 'JHU Applied Physics Laboratory',
                                  Hemisphere.North, 100, [33, -75], [44, 5],
                                  read_hdw_file('wal')),
              RadarID.BPK: _Radar('Buckland Park', 'La Trobe University',
                                  Hemisphere.South, 75, [-30, 138], [-40, -146],
                                  read_hdw_file('bpk')),
              RadarID.DCE: _Radar('Dome C East',
                         'Institute for Space Astrophysics and Planetology',
                                  Hemisphere.South, 75, [-80, 130], [-83, 0],
                                  read_hdw_file('dce')),
              RadarID.DCN: _Radar('Dome C North',
                         'Institute for Space Astrophysics and Planetology',
                                  Hemisphere.South, 75, [-75, 112], [-85, 90],
                                  read_hdw_file('dcn')),
              RadarID.FIR: _Radar('Falkland Islands', 'British Antarctic Survey',
                                  Hemisphere.South, 110, [-47, -59], [-35, 10],
                                  read_hdw_file('fir')),
              RadarID.HAL: _Radar('Halley', 'British Antarctic Survey',
                                  Hemisphere.South, 100, [-71, -27], [-58, 30],
                                  read_hdw_file('hal')),
              RadarID.KER: _Radar('Kerguelen', 'IRAP/CNRS/IPEV', Hemisphere.South,
                                  75, [-44, 70], [-53, 124], read_hdw_file('ker')),
              RadarID.MCM: _Radar('McMurdo', 'University of Alaska, Fairbanks',
                                  Hemisphere.South, 75, [-78, 187], [-75, -36],
                                  read_hdw_file('mcm')),
              RadarID.SAN: _Radar('SANAE', 'South African National Space Agency',
                                  Hemisphere.South, 110, [-67, -3], [-60, 45],
                                  read_hdw_file('san')),
              RadarID.SPS: _Radar('South Pole Station',
                         'University of Alaska, Fairbanks', Hemisphere.South,
                                  75, [-87, 12], [-74, 25], read_hdw_file('sps')),
              RadarID.SYE: _Radar('Syowa East', 'National Institute of Polar Research',
                                  Hemisphere.South, 75, [-64, 45], [-62, 82],
                                  read_hdw_file('sye')),
              RadarID.SYS: _Radar('Syowa South', 'National Institute of Polar Research',
                                  Hemisphere.South, 80, [-66, 30], [-62, 68],
                                  read_hdw_file('sys')),
              RadarID.TIG: _Radar('Tiger', 'La Trobe University', Hemisphere.South,
                                  75, [-38, 147], [-49, -133], read_hdw_file('tig')),
              RadarID.UNW: _Radar('Unwin', 'La Trobe University', Hemisphere.South,
                                  75, [-42, 168], [-49, -105], read_hdw_file('unw')),
              RadarID.ZHO: _Radar('Zhongshan', 'Polar Research Institute of China',
                                  Hemisphere.South, 70, [-67, 64], [-70, 99],
                                  read_hdw_file('zho'))}



# Copyright (C) 2023 SuperDARN Canada, University of Saskatchewan
# Author: Carley Martin
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import bz2
import datetime as dt
import pytest
import warnings

import pydarn


with bz2.open('test/data/test.fitacf.bz2') as fp:
    fitacf_stream = fp.read()
data = pydarn.SuperDARNRead(fitacf_stream, True).read_fitacf()


class TestUtils_citations:
    def test_citations(self):
        with warnings.catch_warnings(record=True):
            pydarn.Citations.print_citations()
    def test_acks(self):
        with warnings.catch_warnings(record=True):
            pydarn.Citations.print_acknowledgements()


class TestUtils_filters:
    def test_boxcar(self):
        with warnings.catch_warnings(record=True):
            bx = pydarn.Boxcar(thresh=0.7, w=None)
            bx.run_filter(data)


class TestUtils_general:
    def test_greatcircle(self):
        with warnings.catch_warnings(record=True):
            pydarn.GeneralUtils.great_circle(100, 50, 110, 60)
    def test_newcoord(self):
        with warnings.catch_warnings(record=True):
            pydarn.GeneralUtils.new_coordinate(100, 50, 100, 30)


@pytest.mark.parametrize('tdiff', [0.003, -0.003])
@pytest.mark.parametrize('overwrite', [True, False])
@pytest.mark.parametrize('interferometer_offset', [[0.0, 100.0, 0.0],
                                                  [1.0, -10.0, 2.0]])
class TestUtils_tdiff:
    def test_recalcelv(self, tdiff, overwrite, interferometer_offset):
        with warnings.catch_warnings(record=True):
            pydarn.recalculate_elevation(data, tdiff=tdiff,
                                         overwrite=overwrite,
                                         interferometer_offset=
                                         interferometer_offset)


class TestUtils_terminator:
    def test_terminator(self):
        with warnings.catch_warnings(record=True):
            pydarn.terminator(dt.datetime(2023,10,10,1,30), 300)


class TestUtils_calcazi:
    def test_calculateazimuth(self):
        with warnings.catch_warnings(record=True):
            pydarn.calculate_azimuth(100, 50, 100, 110, 60, 100)


# Copyright 2019 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modification:
# 2022-03-10 MTS removed gate2slant and gate2groundscatter
#                to range_estimations.py
"""
This module is to focus on data conversions

All methods in this module are functions for
quick conversions.
"""
from typing import List
import numpy as np
from collections import OrderedDict


# key is the format char type defined by python,
# item is the DMAP int value for the type
DMAP_FORMAT_TYPES = {'c': 1,  # char = int8 by RST rtypes.h definition
                     'h': 2,  # short integer
                     'i': 3,  # integer (int)
                     'f': 4,  # float
                     'd': 8,  # double
                     's': 9,  # string
                     'q': 10,  # Long int
                     'B': 16,  # unsigned char
                     'H': 17,  # unsigned short int
                     'I': 18,  # unsigned int
                     'Q': 19}  # unsigned long int

# python possible types that could be the python format type characters
# python doesn't have char types, instead it converts them to strings.
# This conversion is okay because RST treats chars as only int8 types.
FORMAT_CONVERSION = {np.int8: 'c',  # RST defines char as an int8 in rtypes.h
                     np.int16: 'h',
                     int: 'i',
                     np.int32: 'i',
                     float: 'f',
                     np.float32: 'f',
                     np.float64: 'd',
                     str: 's',
                     np.int64: 'q',
                     np.uint8: 'B',
                     np.uint16: 'H',
                     np.uint32: 'I',
                     np.uint64: 'Q'}

DMAP_CASTING_TYPES = {'c': np.int8,  # RST defined char
                      'h': np.int16,  # Short
                      'i': int,  # int
                      'f': float,  # Float
                      'd': np.float64,  # Double
                      's': str,  # String
                      'q': np.int64,  # long int
                      'B': np.uint8,  # Unsigned char
                      'H': np.uint16,  # Unsigned short
                      'I': np.uint32,  # Unsigned int
                      'Q': np.uint64}  # Unsigned long int


def dmap2dict(dmap_records: List[dict]) -> List[dict]:
    """
    This method converts dmap records containing dmap data structures
    to a list of dictionaries.

    Parameters
    ----------
    dmap_records : List[dict]
        a list of dmap records contain dmap data structures

    Return
    ------
    dmap_dict : List[dict]
        a list of dictionaries containing the name of the fields in the keys
        the data value(s) in the items of the dictionary
    """
    dmap_list = []
    for dmap_record in dmap_records:
        dmap_dict = {field: (data.value if isinstance(data.value, np.ndarray)
                     else DMAP_CASTING_TYPES[data.data_type_fmt](data.value))
                     for field, data in dmap_record.items()}
        dmap_list.append(OrderedDict(dmap_dict))
    return dmap_list



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Carley Martin
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import matplotlib.pyplot as plt
import pytest
import warnings

import pydarn


data = pydarn.SuperDARNRead('test/data/test.north.mp').read_map()


class TestMap_defaults:

    def test_map_defaults(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.Maps.plot_mapdata(data)

@pytest.mark.parametrize('colorbar', [False])
@pytest.mark.parametrize('colorbar_label', 'green')
@pytest.mark.parametrize('title', [False])
@pytest.mark.parametrize('cmap', [plt.get_cmap('rainbow')])
@pytest.mark.parametrize('time_delta', [2])
@pytest.mark.parametrize('radar_location', [True])
@pytest.mark.parametrize('zmin', [0])
@pytest.mark.parametrize('zmax', [100])
@pytest.mark.parametrize('len_factor', [150.0, 100.0])
@pytest.mark.parametrize('color_vectors', [False])
@pytest.mark.parametrize('hmb', [False])
@pytest.mark.parametrize('map_info', [False])
@pytest.mark.parametrize('imf_dial', [False])
@pytest.mark.parametrize('reference_vector', [1000])
class TestMap:

    def test_plot_map(self, colorbar, colorbar_label,
                      title, cmap, time_delta, radar_location,
                      zmin, zmax, len_factor, color_vectors, hmb, map_info,
                      imf_dial, reference_vector):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.Maps.plot_mapdata(data, colorbar=colorbar, cmap=cmap,
                                  colorbar_label=colorbar_label, title=title,
                                  time_delta=time_delta,
                                  radar_location=radar_location, zmin=zmin,
                                  zmax=zmax, len_factor=len_factor,
                                  color_vectors=color_vectors, hmb=hmb,
                                  imf_dial=imf_dial,
                                  reference_vector=reference_vector)
        plt.close('all')



# Copyright (C) SuperDARN Canada, University of Saskatchewan
# Authors: Marina Schmidt
#
# Modifications:
# 20210909: CJM - Added NoChannelError
# 20220308 MTS - Added PartialRecordsError()
# 2022-03-23 MTS - Added NotImplementedError() indicates when something is not implement in pyDARN
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import logging
import datetime

pydarn_log = logging.getLogger('pydarn')


class PartialRecordsError(Exception):
    """
    Error given when a partial record prevents a plotting algorithm to work
    """
    def __init__(self, missing_field):
        self.message = "pyDARN could not plot the following due to the "\
                "following field {} was missing. This may be due to partial"\
                "records created in RST (data processing software for"\
                " SuperDARN)".format(missing_field)
        super().__init__(self.message)
        pydarn_log.error(self.message)

class NotImplemented(Exception):
    """
    Error given when settings are not implemented right now.
    """
    def __init__(self, msg):
        self.message = msg
        super().__init__(self.message)
        pydarn_log.error(self.message)

class IncorrectPlotMethodError(Exception):
    """
    This error is raised when the incorrect plotting
    method is called for a given parameter
    """
    def __init__(self, parameter: str, data_type: str):
        self.parameter = parameter
        self.data_type = data_type
        self.message = "Incorrect plotting method is being called for"\
            " {parameter} of type {data_type}. This plot method is for"\
            " array data types, please look at scalar plotting methods "\
            " for this data types.".format(parameter=self.parameter,
                                           data_type=self.data_type)
        super().__init__(self.message)
        pydarn_log.error(self.message)


class IncorrectDateError(Exception):
    """
    This error is raised when the file dates and plotting
    date do not match
    """
    def __init__(self, record_date: datetime.datetime,
                 plot_date: datetime.datetime):
        self.record_date = record_date
        self.plot_date = plot_date
        self.message = "Incorrect date in record {record_date} or "\
            "plotting date {plot_date}"\
            "".format(record_date=self.record_date.strftime('%Y%m%d %H:%M'),
                      plot_date=self.plot_date.strftime('%Y%m%d %H:%M'))
        super().__init__(self.message)
        pydarn_log.error(self.message)


class NoDataFoundError(Exception):
    """
    This error is raised when no data is found for
    the given beam and parameter
    """
    def __init__(self, parameter: str, beam_num: int = None,
                 opt_beam_num: int = None,
                 opt_parameter_value: int = None,
                 start_time: datetime.datetime = None,
                 end_time: datetime.datetime = None):
        self.parameter = parameter
        self.beam_num = beam_num
        self.opt_beam_num = opt_beam_num
        if start_time is None or end_time is None:
            if start_time is not None and beam_num is None:
                self.parameter = parameter
                self.start_time = start_time
                self.message = "There is no record with the start time:"\
                    " {starttime} and parameter {param}. Try another"\
                    " start time, parameter, or increase time_delta"\
                    " option".format(starttime=self.start_time.strftime("%Y"
                                                                        "%m"
                                                                        " %d"
                                                                        " %H:"
                                                                        "%M"),
                                     param=self.parameter)
            elif opt_parameter_value is None:
                self.message = "There is no Data for beam number {beam_num}"\
                        " for the parameter type {parameter}. "\
                        "Try beam, for example: {opt_beam} or"\
                        " another parameter for the given"\
                        " time range.".format(parameter=self.parameter,
                                              beam_num=self.beam_num,
                                              opt_beam=opt_beam_num)
            else:
                self.message = "There is no Data for the beam number"\
                        " {beam_num} for the parameter {parameter}. Try"\
                        " beam {opt_beam} or {parameter} at"\
                        " {parameter_value}"\
                        "".format(beam_num=self.beam_num,
                                  parameter=self.parameter,
                                  opt_beam=self.opt_beam_num,
                                  parameter_value=opt_parameter_value)
        else:
            if opt_parameter_value is None:
                self.start_time = start_time
                self.end_time = end_time
                self.message = "There is no Data for beam number {beam_num}"\
                    " for the parameter type {parameter} between the"\
                    " time range {start_time} to"\
                    " {end_time}. Try beam, for example: {opt_beam}"\
                    " or another parameter for the given time range."\
                    "".format(parameter=self.parameter,
                              beam_num=self.beam_num,
                              opt_beam=opt_beam_num,
                              start_time=self.start_time.strftime("%Y %m"
                                                                  " %d %H"
                                                                  ":%M"),
                              end_time=self.end_time.strftime("%Y %m"
                                                              " %d %H:%M"))
            else:
                self.message = "There is no Data for the beam number"\
                        " {beam_num} for the parameter {parameter}. Try"\
                        " another beam {opt_beam} or {parameter} at"\
                        " {parameter_value} between the time range"\
                        " {start_time} to"\
                        " {end_time}. Try beam, "\
                        "for example: {opt_beam} or another parameter"\
                        " for the given time range."\
                        "".format(beam_num=self.beam_num,
                                  parameter=self.parameter,
                                  opt_beam=self.opt_beam_num,
                                  start_time=self.start_time.strftime("%Y %m"
                                                                      " %d %H"
                                                                      ":%M"),
                                  end_time=self.end_time.strftime("%Y %m"
                                                                  " %d %H:%M"),
                                  parameter_value=opt_parameter_value)

        super().__init__(self.message)
        pydarn_log.error(self.message)


class UnknownParameterError(Exception):
    """
    Error raised when the parameter is not found in the data passed in
    """
    def __init__(self, parameter: str, grid: bool = False):
        """
        parameters
            parameter: KeyError parameter that is not found in the data
            grid: indicating if its called within grid plot or map plot
            that may not have parameters due to how the file was processed
        """
        self.parameter = parameter
        if grid:
            self.message = "The following parameter {parameter}"\
                    " was not found in the"\
                    " data set. Please make sure you used -xtd"\
                    " if you are trying to plot pwd or wdt"\
                    "".format(parameter=self.parameter)
        else:
            self.message = "The following parameter {parameter}"\
                    " was not found in the"\
                    " data set. Please make sure it is typed correctly or"\
                    " you are using the correct data."\
                    "".format(parameter=self.parameter)
        super().__init__(self.message)
        pydarn_log.error(self.message)


class OutOfRangeGateError(Exception):
    """
    This error is raise when range gate requested is out of range.
    """
    def __init__(self, parameter: str, gate_num: int, max_range_gate: int):
        self.parameter = parameter
        self.gate_num = gate_num
        self.max_range_gate = max_range_gate
        self.message = "The range gate {gate_num} is out of range for this"\
            " parameter {param}. Please pick a range gate number"\
            " between 0 - {max_gate}".format(gate_num=self.gate_num,
                                             param=self.parameter,
                                             max_gate=self.max_range_gate-1)
        super().__init__(self.message)
        pydarn_log.error(self.message)


class NoChannelError(Exception):
    """
    This error is raised when the channel specified by user
    does not exist
    """
    def __init__(self, channel: int, opt_channel: int = None):
        self.channel = channel
        self.opt_channel = opt_channel
        self.message = "There is no data for channel {channel},"\
                       " try channel {opt_channel}."\
                       .format(channel=self.channel,
                               opt_channel=self.opt_channel)
        super().__init__(self.message)
        pydarn_log.error(self.message)



# Copyright (C) 2020 SuperDARN Canada, University of Saskatchewan
# Author: Marina Schmidt
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.

import bz2
import datetime as dt
import matplotlib.pyplot as plt
import pytest
import warnings

import pydarn

with bz2.open('test/data/test.fitacf.bz2') as fp:
    fitacf_stream = fp.read()
data = pydarn.SuperDARNRead(fitacf_stream, True).read_fitacf()


class TestFan_defaults:

    def test_fan_defaults(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.Fan.plot_fan(data)

    def test_fov_series(self):
        """ """
        with warnings.catch_warnings(record=True):
            pydarn.Fan.plot_fov(pydarn.RadarID.PGR, dt.datetime(2020, 4, 4, 6, 2))


@pytest.mark.parametrize('stid', [pydarn.RadarID.SAS, pydarn.RadarID.DCN])
@pytest.mark.parametrize('ranges', [(5,70)])
@pytest.mark.parametrize('boundary', [False])
@pytest.mark.parametrize('rsep', [15])
@pytest.mark.parametrize('frang', [90])
@pytest.mark.parametrize('fov_color', ['grey'])
@pytest.mark.parametrize('alpha', [0.8, 1])
@pytest.mark.parametrize('radar_location', [False])
@pytest.mark.parametrize('radar_label', [True])
@pytest.mark.parametrize('line_color', ['red'])
@pytest.mark.parametrize('line_alpha', [0.8, 1])
@pytest.mark.parametrize('date', [dt.datetime(2020, 4, 4, 6, 2)])
@pytest.mark.parametrize('grid', [True, False])
@pytest.mark.parametrize('beam', [None, 7])
class TestFov:

    def test_fov_plot(self, stid, ranges, boundary, fov_color,
                                   alpha, radar_location, radar_label,
                                   line_color, date, grid, line_alpha, rsep,
                                   frang, beam):
        """ this test will give bare minimum input needed for """
        with warnings.catch_warnings(record=True):
            pydarn.Fan.plot_fov(stid=stid, ranges=ranges,
                                boundary=boundary, fov_color=fov_color,
                                alpha=alpha, rsep=rsep, frang=frang,
                                radar_location=radar_location,
                                radar_label=radar_label, grid=grid,
                                line_color=line_color, date=date,
                                line_alpha=line_alpha, beam=beam)
        plt.close('all')


@pytest.mark.parametrize('parameter', ['p_l', 'w_l', 'v'])
@pytest.mark.parametrize('cmap', [plt.get_cmap('rainbow')])
@pytest.mark.parametrize('groundscatter', [True])
@pytest.mark.parametrize('colorbar', [False])
@pytest.mark.parametrize('colorbar_label', ['stringhere'])
@pytest.mark.parametrize('zmin', [0])
@pytest.mark.parametrize('zmax', [100])
@pytest.mark.parametrize('title', [False])
@pytest.mark.parametrize('channel', [1,2])
@pytest.mark.parametrize('ball_and_stick', [True, False])
@pytest.mark.parametrize('len_factor', [300, 100])
@pytest.mark.parametrize('beam', [None, 7])
class TestFan:

    def test_fan_plot(self, parameter, cmap, groundscatter, colorbar,
                      colorbar_label, zmin, zmax, title, channel, beam,
                      ball_and_stick, len_factor):
        """

        """
        with warnings.catch_warnings(record=True):
            pydarn.Fan.plot_fan(data, parameter=parameter, cmap=cmap,
                                groundscatter=groundscatter, colorbar=colorbar,
                                colorbar_label=colorbar_label, zmin=zmin,
                                zmax=zmax, title=title, channel=channel,
                                beam=beam, ball_and_stick=ball_and_stick,
                                len_factor=len_factor)



# Copyright (C) 2023 SuperDARN Canada, University of Saskwatchewan
# Author: Carley Martin
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#

"""
IQ plots
"""
import matplotlib.pyplot as plt
import numpy as np
import warnings

from datetime import datetime
from typing import List

from pydarn import (time2datetime, standard_warning_format, find_record, RTP)

warnings.formatwarning = standard_warning_format


class IQ:
    """
    Time-series and specialised plots for IQ data

    Class pattern design: Builder
    This class inherits matplotlib.pyplot to inherit plotting features as well
    build off their builder design pattern.

    Notes
    -----
    This class inherits from matplotlib to generate the figures

    Methods
    -------
    plot_time_series
    plot_iq_sequence
    plot_iq_record
    plot_iq_overview
    """

    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - plot_time_series()\n"\
                "   - plot_iq_sequence()\n"\
                "   - plot_iq_record()\n"\
                "   - plot_iq_overview()\n"

    @staticmethod
    def plot_time_series(dmap_data: List[dict], **kwargs):
        """
        Plots scalar variables from IQ dat files

        SEE ALSO: RTP.plot_time_series() for more information and
        inputs

        Parameters
        ----------
        dmap_data: List[dict]

        Returns
        -------
        time_series_data: Lines, x, y
            standard return values from RTP.plot_time_series
        """
        time_series_info = RTP.plot_time_series(dmap_data, **kwargs)
        return time_series_info

    @staticmethod
    def plot_iq_sequence(dmap_data: List[dict],
                         start_time: datetime = None,
                         channel: int = 0, ax=None, beam_num: int = 0,
                         sequence_num: int = 0, interferometer: bool = False,
                         plot_phase: bool = False):
        """
        Plots a single sequence from a record of IQ data

        Parameters
        -----------
        dmap_data: List[dict]
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The frequency channel 0, 1, 2 ...
            Default : 0
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        start_time: datetime
            Specified time of interest as a datetime object
        sequence_num: int
            Sequence in the record to be plotted
            Default: 0
        interferometer: bool
            Data from main array (False) or interferometer array (True)
            Default False
        plot_phase: bool
            Option to plot the phase of the data (True)
            Default False
        """
        # Finds start of minute scan records, then scans the records to
        # find the next record with the correct beam
        if start_time is not None:
            record_num = find_record(dmap_data, start_time)
        else:
            record_num = 0
        # Get the next record with the correct beam
        while (dmap_data[record_num]['bmnum'] != beam_num
                and (dmap_data[record_num]['channel'] != channel
                     and channel != 'all')):
            record_num += 1
            if record_num >= len(dmap_data):
                raise Exception("No matching data found for beam {} "
                                "near time {}"
                                .format(beam_num,
                                        start_time.strftime('%Y-%m-%d' +
                                                            ' %H:%M:%S')))
        dmap_record = dmap_data[record_num]

        # Details from chosen record
        date = time2datetime(dmap_record)
        smpnum = dmap_record['smpnum']
        chnnum = dmap_record['chnnum']
        seqnum = dmap_record['seqnum']

        # Warning to too hihc sequence numberes
        if sequence_num >= seqnum:
            raise ValueError("Sequence numbers available for this record "
                             "range from 0 to {}.".format(seqnum-1))

        if interferometer and chnnum < 2:
            raise ValueError("No interferometer data for record chosen.")

        # Calculate starting data position
        # For layout of data and how to access see:
        # https://radar-software-toolkit-rst.readthedocs.io
        # /en/latest/references/general/iqdat/
        if interferometer:
            smp = sequence_num * chnnum * 2 * smpnum + 2 * smpnum
        else:
            smp = sequence_num * chnnum * 2 * smpnum
        iq_real = []
        iq_imag = []
        smp_initial = smp
        while smp < smp_initial + 2*smpnum:
            iq_real.append(dmap_record['data'][smp])
            smp = smp + 1
            iq_imag.append(dmap_record['data'][smp])
            smp = smp + 1

        # Calculate magnitude and phase
        mag = [np.sqrt(i**2 + j**2) for i, j in zip(iq_real, iq_imag)]
        mag_neg = [-np.sqrt(i**2 + j**2) for i, j in zip(iq_real, iq_imag)]

        # Plotting phase if chosen
        if plot_phase:
            phase = np.arctan2(iq_imag, iq_real)
            fig = plt.gcf()
            fig.tight_layout()
            gs = fig.add_gridspec(3, 1)
            ax = fig.add_subplot(gs[0:2, :])
            ax2 = fig.add_subplot(gs[2, :])
            ax2.plot(phase, 'mediumseagreen')
            ax2.grid()
            ax2.set_ylabel('Phase')
            ax2.set_xlabel('Sample Number')
        elif ax is None:
            ax = plt.gca()

        # Plot lines
        real_line, = ax.plot(iq_real, 'r')
        imag_line, = ax.plot(iq_imag, 'b')
        mag_line, = ax.plot(mag, 'grey', linestyle='--', linewidth=0.5)
        ax.plot(mag_neg, 'grey', linestyle='--', linewidth=0.5)

        # Plot legend and labels
        real_line.set_label('Real')
        imag_line.set_label('Imaginary')
        mag_line.set_label('Magnitude')
        ax.legend()

        # Plot title and axis labels
        if interferometer:
            array_type = 'Interferometer Array'
        else:
            array_type = 'Main Array'
        ax.set_title('IQ Data for ' + array_type + '\n Beam='
                     + str(beam_num) + ' '
                     + ' Sequence=' + str(sequence_num) + '\n '
                     + date.strftime('%Y-%m-%d %H:%M:%S'))
        ax.set_ylabel('Raw Data')
        ax.set_xlabel('Sample Number')
        ax.grid()

        return {'ax': ax,
                'ccrs': None,
                'cm': None,
                'cb': None,
                'fig': plt.gcf(),
                'data': {'iq_real': iq_real,
                         'iq_imag': iq_imag,
                         'magnitude': mag}
                }

    @staticmethod
    def plot_iq_record(dmap_data: List[dict],
                       start_time: datetime = None,
                       channel: int = 0, ax=None, beam_num: int = 0,
                       interferometer: bool = False,
                       cmap=plt.colormaps['viridis']):
        """
        Plots all sequences from a record of IQ data

        Parameters
        -----------
        dmap_data: List[dict]
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The frequency channel 0, 1, 2 ...
            Default : 0
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        start_time: datetime
            Specified time of interest as a datetime object
        interferometer: bool
            Data from main array (False) or interferometer array (True)
            Default False
        cmap: matplotlib.cm object
            Default: viridis
        """
        # Finds start of minute scan records, then scans the records to
        # find the next record with the correct beam
        if start_time is not None:
            record_num = find_record(dmap_data, start_time)
        else:
            record_num = 0
        # Get the next record with the correct beam
        while (dmap_data[record_num]['bmnum'] != beam_num
                and (dmap_data[record_num]['channel'] != channel
                     and channel != 'all')):
            record_num += 1
            if record_num >= len(dmap_data):
                raise Exception("No matching data found for beam {} "
                                "near time {}"
                                .format(beam_num,
                                        start_time.strftime('%Y-%m-%d' +
                                                            ' %H:%M:%S')))
        dmap_record = dmap_data[record_num]

        # Details from chosen record
        date = time2datetime(dmap_record)
        smpnum = dmap_record['smpnum']
        chnnum = dmap_record['chnnum']
        seqnum = dmap_record['seqnum']

        if interferometer and chnnum < 2:
            raise ValueError("No interferometer data for record chosen.")

        # Calculate starting data position
        # For layout of data and how to access see:
        # https://radar-software-toolkit-rst.readthedocs.io
        # /en/latest/references/general/iqdat/
        iq_real_arr = np.empty([smpnum, seqnum])
        iq_imag_arr = np.empty([smpnum, seqnum])
        for seq in range(0, seqnum):
            if interferometer:
                smp = seq * chnnum * 2 * smpnum + 2 * smpnum
            else:
                smp = seq * chnnum * 2 * smpnum
            iq_real_arr[:, seq] = dmap_record['data'][smp: smp + smpnum]
            iq_imag_arr[:, seq] =\
                dmap_record['data'][smp + smpnum: smp + 2*smpnum]

        # Calculate magnitude
        mag = np.sqrt(iq_real_arr**2 + iq_imag_arr**2)
        # Plot
        if ax is None:
            ax = plt.gca()
        pcol = ax.pcolormesh(mag, vmin=0, vmax=600, cmap=cmap)
        cb = ax.figure.colorbar(pcol, extend='max')
        cb.set_label('Power (AU)')

        # Plot title and axis labels
        if interferometer:
            array_type = 'Interferometer Array'
        else:
            array_type = 'Main Array'
        ax.set_title('IQ Data for ' + array_type + '\n Beam='
                     + str(beam_num) + ' '
                     + date.strftime('%Y-%m-%d %H:%M:%S'))
        ax.set_ylabel('Sample Number')
        ax.set_xlabel('Sequence Number')
        ax.grid()

        return {'ax': ax,
                'ccrs': None,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'iq_real': iq_real_arr,
                         'iq_imag': iq_imag_arr,
                         'magnitude': mag}
                }

    @staticmethod
    def __determine_start_end_time(dmap_data, start_time: datetime,
                                   end_time: datetime) -> tuple:
        """
        Sets the start and end time based on import of dmap_data

        Parameter
        ---------
        dmap_data: List[dict]
        start_time: datetime
            Start time is used to check if it was set or not
        end_time: datetime
            End time is used to check if it was set or not

        Returns
        -------
        start_time: datetime
        end_time: datetime
        """
        if not start_time:
            start_time = time2datetime(dmap_data[0])
        if not end_time:
            end_time = time2datetime(dmap_data[-1])
        return start_time, end_time

    @staticmethod
    def plot_iq_overview(dmap_data: List[dict],
                         start_time: datetime = None,
                         end_time: datetime = None,
                         channel: int = 0, ax=None, beam_num: int = 'all',
                         interferometer: bool = False,
                         cmap: object = plt.colormaps['viridis']):
        """
        Plots all sequences from a record of IQ data

        Parameters
        -----------
        dmap_data: List[dict]
        beam_num : int
            The beam number of data to plot
            Default: 0
        channel : int or str
            The frequency channel 0, 1, 2 ...
            Default : 0
        ax: matplotlib.axes
            axes object for another way of plotting
            Default: None
        start_time: datetime
            Specified start time of interest as a datetime object
        end_time: datetime
            Specified end time of interest as a datetime object
        interferometer: bool
            Data from main array (False) or interferometer array (True)
            Default False
        cmap: matplotlib.cm object
            Default: viridis
        """
        start_time, end_time = IQ.__determine_start_end_time(dmap_data,
                                                             start_time,
                                                             end_time)

        # Assuming the samples and sequences will stay constant
        # over the file
        smpnum = dmap_data[0]['smpnum']
        chnnum = dmap_data[0]['chnnum']
        seqnum = dmap_data[0]['seqnum']

        # Empty lists to append data to
        # TODO: did not work using arrays
        iq_real_arr = []
        iq_imag_arr = []
        for dmap_record in dmap_data:
            if (dmap_record['bmnum'] == beam_num or beam_num == 'all') and\
                    (dmap_record['channel'] == channel or channel == 'all'):
                rec_time = time2datetime(dmap_record)
                if start_time <= rec_time <= end_time:
                    if interferometer and chnnum < 2:
                        raise ValueError("No interferometer data for "
                                         "record chosen.")

                    # Calculate starting data position
                    # For layout of data and how to access see:
                    # https://radar-software-toolkit-rst.readthedocs.io
                    # /en/latest/references/general/iqdat/
                    for seq in range(0, seqnum):
                        if interferometer:
                            smp = seq * chnnum * 2 * smpnum + 2 * smpnum
                        else:
                            smp = seq * chnnum * 2 * smpnum
                        iq_real = dmap_record['data'][smp: smp + smpnum]
                        iq_imag =\
                            dmap_record['data'][smp + smpnum: smp + 2*smpnum]
                        if len(iq_real) == smpnum:
                            iq_real_arr.append(iq_real)
                            iq_imag_arr.append(iq_imag)

        # Convert to arrays, dtype important for mag calc
        iq_real_arr = np.array(iq_real_arr, dtype='float64')
        iq_imag_arr = np.array(iq_imag_arr, dtype='float64')
        # Calculate magnitude
        mag = np.sqrt(np.add(np.square(iq_real_arr), np.square(iq_imag_arr)))

        # Plot
        if ax is None:
            ax = plt.gca()
        pcol = ax.pcolormesh(mag.T, vmin=0, vmax=1000, cmap=cmap)
        cb = ax.figure.colorbar(pcol, extend='max')
        cb.set_label('Power (AU)')

        # Plot title and axis labels
        if interferometer:
            array_type = 'Interferometer Array'
        else:
            array_type = 'Main Array'
        ax.set_title('IQ Data for ' + array_type + ' Beam='
                     + str(beam_num) + '\n'
                     + start_time.strftime('%Y-%m-%d %H:%M:%S') + ' - '
                     + end_time.strftime('%Y-%m-%d %H:%M:%S'))
        ax.set_ylabel('Sample Number')
        ax.set_xlabel('Sequence Number')
        ax.grid()

        return {'ax': ax,
                'ccrs': None,
                'cm': cmap,
                'cb': cb,
                'fig': plt.gcf(),
                'data': {'iq_real': iq_real_arr,
                         'iq_imag': iq_imag_arr,
                         'magnitude': mag}
                }



# Copyright (C) 2023 SuperDARN Canada, University of Saskatchewan
# Author(s): Carley Martin
#
# Modifications:
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
"""
This module contains general calculations used across pyDARN library and may be
useful for the user
"""
import numpy as np

from pydarn import Re


class GeneralUtils():
    """
        General Utility Functions

        Methods
        -------
        great_circle
        new_coordinate
    """
    def __str__(self):
        return "This class is static class that provides"\
                " the following methods: \n"\
                "   - great_circle()\n"

    @classmethod
    def great_circle(cls, lon1, lat1, lon2, lat2):
        '''
        Calculates the great circle distance between two points on a sphere

        Parameters
        -----------
            lon1: float in degrees
                longitude of first point
            lat1: float in degrees
                latitude of first point
            lon2: float in degrees
                longitude of second point
            lat1: float in degrees
                latitude of second point

        Returns
        -------
            distance: float
                assuming radius of sphere is 1 (1 Re)
        '''
        lon1 = np.radians(lon1)
        lat1 = np.radians(lat1)
        lon2 = np.radians(lon2)
        lat2 = np.radians(lat2)
        return (np.arccos(np.sin(lat1) * np.sin(lat2)
                          + np.cos(lat1) * np.cos(lat2)
                          * np.cos(lon1 - lon2)))

    @classmethod
    def new_coordinate(cls, lat, lon, d, bearing, R=Re):
        """
        new coordinate will calculate the new coordinate from a given
        position, distance and bearing

        Parameters
        ----------
        lat: float
            initial latitude, in degrees
        lon: float
            initial longitude, in degrees
        d: target distance from initial
        bearing: (true) heading in degrees
        R: optional radius of sphere, defaults to mean radius of earth

        Returns
        -------
        lat: float
            new latitude, in degrees
        lon: float
            new longitude, in degrees
        """
        lat1 = np.radians(lat)
        lon1 = np.radians(lon)
        a = np.radians(bearing)
        lat2 = np.arcsin(np.sin(lat1) * np.cos(d/R) + np.cos(lat1)
                         * np.sin(d/R) * np.cos(a))
        lon2 = lon1 + np.arctan2(np.sin(a) * np.sin(d/R) * np.cos(lat1),
                                 np.cos(d/R) - np.sin(lat1) * np.sin(lat2))
        return (np.degrees(lat2), np.degrees(lon2))



# Copyright (C) 2020 SuperDARN Canada, Universtiy of Saskatchewan
# Author(s): Daniel Billett
#
# Disclaimer:
# pyDARN is under the LGPL v3 license found in the root directory LICENSE.md
# Everyone is permitted to copy and distribute verbatim copies of this license
# document, but changing it is not allowed.
#
# This version of the GNU Lesser General Public License incorporates the terms
# and conditions of version 3 of the GNU General Public License,
# supplemented by the additional permissions listed below.
#
# Modification:
#
"""
This module is used for sorting a given dmap_data list of dictionaries
into scans
"""

import datetime
import numpy as np
from typing import List
from pydarn import time2datetime


def build_scan(dmap_data: List[dict]):
    """
    Returns list of size equal to number of records in dmap_data,
    with scan number for each record

    Parameters
    ----------
    dmap_data: List(dict)
        list of records (dictionaries) representing dmap data
    Returns
    ----------
    beam_scan: List
        list of size equal to number of records in dmap_data, with scan number
    for each record
    """
    # Set up scans for easy locating
    # Makes a list of size (number of records), with the scan number for each
    scan_mark = [sub['scan'] for sub in dmap_data]
    timestamps = [time2datetime(rec) for rec in dmap_data]
    timestamps_seen = []
    current_scan = 0
    first_time = True
    beam_scan = np.zeros((len(dmap_data)))
    for i in range(len(dmap_data)):
        # Check if this record is concurrent with another record
        if timestamps[i] in timestamps_seen:
            beam_scan[i] = current_scan
        else:
            timestamps_seen.append(timestamps[i])   # Record the timestamp from this file

            # Absolute value used due to some scan flags set as "-1"
            if abs(scan_mark[i]) == 1 and not first_time:
                current_scan += 1
            first_time = False
            beam_scan[i] = current_scan

    return beam_scan


def find_records_by_datetime(dmap_data: List[dict], search: datetime.datetime,
                             tolerance: datetime.timedelta):
    """
    Returns the records which are within a certain timeframe relative to a search time.

    Parameters
    ----------
    dmap_data: List(dict)
        list of records (dictionaries) representing dmap data
    search: datetime.datetime
        datetime to search for records matching this time
    tolerance: datetime.timedelta
        how far off in time the records can be and still be a hit
    Returns
    ----------
    recs: List(dict)
        list of records that are close enough in time to search value
    """
    timestamps = np.array([time2datetime(rec) for rec in dmap_data])
    matches = np.nonzero(np.abs(timestamps - search) < tolerance)[0]
    return [dmap_data[match] for match in matches]


def find_records_by_scan(dmap_data: List[dict], scan_index: int):
    """
    Returns the records which are from the scan corresponding to scan_index
    in the file.

    Parameters
    ----------
    dmap_data: List(dict)
        list of records (dictionaries) representing dmap data
    scan_index: int
        index of scan in file to return
    Returns
    ----------
    recs: List(dict)
        list of records that match the search criteria
    """
    scan_indices = build_scan(dmap_data)
    matches = np.nonzero(scan_indices == scan_index)[0]
    return [dmap_data[match] for match in matches]